"""
Code printing out and examing Cluto clusters assuming
cluto's output files are already available.

cluto/dependency_DB_500_NN_key.clu   This is built by code in the Python cluto_prep module (this package)
cluto/dep_500_nn.cluster    cluto output file.  The actual clusters.  each row in the Cluto matrix file is
                            represented by a line.  The line contains the number of the cluster containing that row.
cluto/dependency_DB_500_NN.mat.tree  cluto output file representing entire binary clustering tree (2k -1) lines
                                     where k is the number of rows in the input matrix. Not to be confused
                                      with .cltree file with only 2n -1 lines where n is the number of clusters

For the next commandline to work, the two output files, dependency_DB_500_NN.mat,
and dependency_DB_500_NN.mat.rlabel shd be copied to $cluto/Matrices/.
As written the next command is administered from the $cluto directory (which has binaries for various machine architectures stored
in various subdirectories)::

Darwin-i386/vcluster -rlabelfile=Matrices/dependency_DB_500_NN.mat.rlabel -clustfile=Matrices/dep_500_nn.cluster Matrices/dependency_DB_500_NN.mat 20

I call 'Matrices/dep_500_nn.cluster' the cluster file. It is vclusters report.  There
is also a lot of screen dumped data summarizing quality info for the clusters.

Then assuming this clusters file and the key file, C{dependency_DB_500_NN_key.clu}, generated by default by C{write_cluto_file},
have both been copied to relative path C{cluto} from cwd, call C{get_clusters} as follows.

>>> get_clusters('cluto/dependency_DB_500_NN_key.clu','cluto/dep_500_nn_agglo.cluster', 20)

This prints out the cluster words from the clusters file.


>>> get_tree('cluto/dependency_DB_500_NN_key.clu','cluto/dep_500_nn_agglo.cluster', 50, 'cluto/dependency_DB_500_NN.mat.tree',make_tree_str=True)

This prints out the cluster tree from the clusters file.
"""
from dep_file_utils import Ctr


##  Moved this from cluto_prep.py
def get_tree (key_file, cluster_file, num_clusters, tree_file,make_tree_str=False):
    """
    First gets the C{clusters} and C{lexemes} from C{cluster_file}, a cluto generated cluster file (extension ".cluster"),
    using C{key_file} for labeling.  Then gets the internal hierarchical cluster tree from
    C{tree_file}. Format of C{tree_file}.  THe first space-separated element of row $k$ is
    the index of $k$'s parent.  More or less the top level function in this module.
    """
    clusters, lexemes = get_clusters(key_file, cluster_file, num_clusters)
    tfh = open(tree_file,'r')
    parents = []
    children = {}
    ctr = 0
    for line in tfh:
        parent = int(line.strip().split()[0])
        parents.append(parent)
        if parent in children:
            children[parent].append(ctr)
        else:
            children[parent] = [ctr]
        ctr += 1
    if make_tree_str:
       return make_hierarchy (children, lexemes, len(lexemes))
    return (clusters,lexemes,parents,children)

##  Moved this from cluto_prep.py
def get_clusters (key_file, cluster_file, num_clusters, display=False):
    """
    Just gets the clusters from C{cluster_file}, a cluto generated cluster file (extension ".cluster"),
    Uses C{key_file} for labeling.  This has usually been generated by Python code from the relational
    db component, which in turn is generated from dependency parse trees.
    """

    clusters = {}
    lexemes = []
    for k in range(-1,num_clusters):
        clusters[k] = []
    kfh = open(key_file,'r')
    cfh = open(cluster_file,'r')
    for line in cfh:
        clust = int(line.strip())
        lexeme = kfh.readline().strip()
        lexemes.append(lexeme)
        clusters[clust].append(lexeme)
    kfh.close()
    cfh.close()
    if display:
        print_clusters (clusters)
    return clusters,lexemes

def print_clusters (clusters):
     for k in clusters:
         print 'Cluster %d:' % k
         for lex in clusters[k]:
             print '    %s' % lex    

def get_node_sizes(children,leaf_size):
    tree_size = tree_size_from_leaf_size (leaf_size)
    node_sizes= {}
    get_node_size (tree_size, children,leaf_size,node_sizes)
    return node_sizes

def get_node_size (node, children,leaf_size,node_sizes=None):
   """
   Use this to find out how many words are dominated
   by an internal node of the hierarchical cluster tree.
   """
   if node_sizes is None:
      node_sizes = {}
   if node < leaf_size:
      # leaf nodes have lowest nums. We are AT a leaf node
      node_sizes[node] = 1
   else:
      node_sizes[node] = 0
      for child in children[node]:
         get_node_size(child,children, leaf_size,node_sizes)
         node_sizes[node] += node_sizes[child]
   return node_sizes

def tree_size_from_leaf_size (leaf_size):
    n = (2 * leaf_size) - 3
    print n
    return n


## For the 500 NN case do the following:
## make_hierarchy(children,lexemes,500)
def make_hierarchy (children, lexemes, leaf_size):
   """
   C{children} is dictionary which for each parentr returns the
   list of its children in the hierarchy tree. Usually
   C{children} and C{lexemes} are computed by calling
   C{get_tree} on the the tree and cluster files generated
   by Cluto.
   """
   tree_size = tree_size_from_leaf_size (leaf_size)
   return make_hierarchy_helper (tree_size, children, lexemes, leaf_size)

def make_hierarchy_helper (n, children, lexemes, leaf_size):
   # print n, leaf_size
   if n >= leaf_size:
      (c1, c2) = children[n]
      return '(%s, %s)' % (make_hierarchy_tree(c1,children,lexemes,leaf_size), make_hierarchy_tree(c2,children,lexemes,leaf_size))
   else:
      return lexemes[n].split()[1]

def make_hierarchy_tree (root, children, lexemes,leaf_size):
   """
   Dunno whatthis is for.
   """
   if root < leaf_size:
      return lexemes[root].split()[1]
   else:
      return make_hierarchy_helper(root,children,lexemes,leaf_size)

if __name__ == '__main__':
   # nn_dict = {}
   # read_nn_file('rel_pp_nearest_neighbor_%s.dict' % pos,suffix='_cos',ext='_pmi',nn_dict=nn_dict)
   #stem = 'dependency_DB_500_NN'
   stem = 'militant_data_DB_NN'
   key_file_name = '%s_key.clu' % stem
   tree_file_name = '%s.mat.tree' % stem
   #cluster_file_name = 'dep_500_nn.cluster'
   cluster_file_name = '%s.mat.clustering.50' % stem
   
   (clusters,lexemes,parents,children) = get_tree('cluto/%s' % key_file_name, \
                                                  'cluto/%s' % cluster_file_name, 50, \
                                                  'cluto/%s' % tree_file_name)
   ## To get the stringfor ENTIRE hierarchy to put in the tex file.  NB: Tex capacity exceeded warning.
   ## TeX capacity exceeded, sorry [grouping levels=255]
   ## make_hierarchy called to make H_str
   ##H_str = get_tree('cluto/dependency_DB_500_NN_key.clu',   'cluto/dep_500_nn.cluster', 50, 'cluto/dependency_DB_500_NN.mat.tree',make_tree_str=True)
   H_str = make_hierarchy (children, lexemes, len(lexemes))
   # You can also use get_tree to do this, but it's an obverliaded use of this fn.
   # H_str = get_tree('cluto/%s' % key_file_name, \
   #                 'cluto/%s' % cluster_file_name, 50, \
   #                 'cluto/%s' % tree_file_name,\
   #                 make_tree_str=True)
   ## Useful info for figuring what nodes to print in latex.
   ## get number of leaves dominated by each node in the hierarchy.
   ns = get_node_sizes(children,len(lexemes))
   ## Get a large chunk of the hierarchy
   ## Higher number nodes created later, nodes higher in tree in agglomerative clustering.
   ## Contains body, gun, genocide, murder, fraud, assassination
   H = make_hierarchy_helper (991, children, lexemes, 500)

   
   
