word_ct_dict = {}
# A sorted wd ct item list, made form the above dict.
sorted_word_cts = []
external_ct_dict_file = ''


import re
import os.path, sys
## Access to symbolic indexes into the 10-tuple returned by os.stat()
from stat import *
from shutil import copy2
from heapq import heappush, heappop
import numpy as np


# lex_list = get_vocab(filename,pos,top_n)
# get_vocab('../dep_data/dependency_DB.train','NN',500)
def get_vocab(filename,pos,top_n):
    global word_ct_dict, sorted_wd_cts
    (pos_str,top_n_str) = get_file_suffixes(pos, top_n)
    (base, ext) = os.path.splitext(filename)
    
    ## An items file contains sorted counts, so we dont sort it on read in
    ## This one contains JUST the vocab needed for this experiment
    freq_filename = base + top_n_str + pos_str + '_freq' + '.items'
    if os.path.isfile(freq_filename) and os.stat(freq_filename)[ST_SIZE] > 0:
        return read_ct_file (freq_filename)
    
    ## A global sorted freq file prepared by another program.
    ## Bigger but still saves us the task of sorting.
    if external_ct_dict_file and os.path.isfile(external_ct_dict_file):
        return read_ct_file (freq_filename)
    
    ## Otherwise we compute cts now.
    ## Note:  These word cts are NOT real token cts.  They are dep rel cts:
    ## The big dog barked. => The token 'dog' enters into 3 deprels.
    #word_ct_dict = get_word_counts(filename)
    #sorted_wd_cts = sort_freq_dict(word_ct_dict)
    sorted_wd_cts = get_sorted_word_counts(filename)
    
    ## Save it for next time!
    save_ct_items (sorted_wd_cts,freq_filename,pos,top_n)
    return sorted_wd_cts

def parse_dep_line(line, count_file=False):
    pairs = line.strip().split()
    ## Will raise error for line not matching format, as desired.
    (word,pos) = pairs[0].split('=')
    dep_pairs = [ ]
    for dep_ct_pair in pairs[1:]:
        (feat_str,count_str) = dep_ct_pair.split(':')
        if count_file:
            feat,count = (int(feat_str),int(count_str))
        else:
            feat,count = (int(feat_str),float(count_str))
        dep_pairs.append((feat,count))
    return ((word, pos), dep_pairs)

inverse_deprel_char = '-'
def parse_dep_rel_key_line (line, line_no=0):
    """
    Special purpose function for
    parsing lines of the dependency_DB.keys file.

    NOT To be confused with C{dep_file_utils.parse_dep_line}
    which parses lines of the dependency_DB.train file.
    Both needed.
    """
    error_str = '*** Error ***: Non parsable dep_rel line no %d: %s' % (line_no,line)
    try:
        (wd,dep_rel0) = eval(line)
    except TypeError:
        raise Exception, error_str 
    except ValueError:
        raise Exception, error_str
    dep_elements = dep_rel0.split('_')
    try:
        (pos,dep_rel) = dep_elements
    except ValueError:
        ## more than 2 elements
        pos = dep_elements[0]
        dep_rel = '_'.join(dep_elements[1:])
        if dep_rel:
            pass
        else:
            # only 1 element, no '_'
            if dep_rel0 in  ['ROOT', '-ROOT']:
                pos = 'ROOT'
                dep_rel = dep_rel0
            else:
                pos = ''
                print '*** Warning ***: No part of speech in line no %d dep_rel %s' % (line_no,line)
    if pos.startswith(inverse_deprel_char):
        dep_rel = '-' + dep_rel
        pos = pos[1:]
    return (wd, dep_rel,pos)

#def update_word_ct_dict(lexeme,token_ctr,count_file):
#    global word_ct_dict

def update_ct_dict (dd,key,increment):
    if key in dd:
        dd[key] += increment
    else:
        dd[key] = increment


def sort_freq_dict (fdd,reverse=True,sort_index=1,check_count=True):
    """
    We're (perhaps unwisely) using count_dicts to represent sets
    sometimes, so the values may not be counts.  Convention is
    to set all vals to -1 in this case.  Test for this first.

    sort_index=0 allows sorting by keys.
    """
    if check_count and fdd.itervalues().next() < 0:
        raise Exception, 'Count dictionary has not been initialized with counts'
    print 'sorting freq dict'
    il = fdd.items()
    il.sort(key = lambda x: x[sort_index],reverse=reverse)
    print 'sorting completed'
    return il

# word_ct_dict = get_word_counts(filename)
def get_word_counts (filename):
    fh = open(filename,'r')
    print 'Getting word counts'
    word_ct_dict = { }
    for line in fh:
        try:
            (lexeme, dep_ct_pairs) = parse_dep_line(line, count_file=True)
        except TypeError:
            continue
        except ValueError:
            continue
        ctr = 0
        if dep_ct_pairs:
            for (feat,count) in dep_ct_pairs:
                ctr += count
        else:
            ctr += 1
        word_ct_dict[lexeme] = ctr
    fh.close()
    print 'Words counted'
    return word_ct_dict

def get_sorted_word_counts (filename):
    """
    Python heaps are min heaps.

    So we negate the scores to get most frequent words
    at the top.
    """
    fh = open(filename,'r')
    print 'Getting word counts'
    # word_ct_dict = { }
    il = []
    for line in fh:
        try:
            (lexeme, dep_ct_pairs) = parse_dep_line(line, count_file=True)
        except TypeError:
            continue
        except ValueError:
            continue
        ctr = 0
        if dep_ct_pairs:
            for (feat,count) in dep_ct_pairs:
                ctr += count
        else:
            ctr += 1
        heappush(il, (-ctr, lexeme))
        # word_ct_dict[lexeme] = ctr
    fh.close()
    print 'Words counted'
    ##  Now collect results in C{res}, remembering to reverse polarity of negative cts.
    res = []
    while il:
        (ct, lex) = heappop(il)
        res.append((lex, - ct))
    return res

def sort_word_deprel_dict (word_deprel_dict):
    """
    Python heaps are min heaps.

    So we negate the scores to get most frequent words
    at the top.
    """
    il = []
    print 'Counting deprels'
    for (lexeme, dep_ct_pairs) in word_deprel_dict.iteritems():
        ctr = 0
        if dep_ct_pairs:
            for (feat,count) in dep_ct_pairs.iteritems():
                ctr += count
        else:
            ctr += 1
        heappush(il, (-ctr, lexeme))
        # word_ct_dict[lexeme] = ctr
    print 'Deprels counted'
    ##  Now collect results in C{res}, remembering to reverse polarity of negative cts.
    res = []
    while il:
        (ct, lex) = heappop(il)
        res.append((lex, - ct))
    return res

def save_ct_items (il,filename,pos,top_n):
    """
    Parameter C{il} is a sorted item list.  We save top n (C{top_n}) items
    matching C{pos} in C{filename}.
    """
    fh = open(filename,'w')
    ctr = 0
    if type(top_n) == tuple:
        pass
    else:
        top_n = (0,top_n)
    # Cant just take a splice of C{il}
    # because of C{pos} filtering.
    for ((word_i,pos_i), v) in il:
        if pos and pos_i <> pos:
            continue
        ctr += 1
        if ctr > top_n[1]:
            break
        elif ctr < top_n[0]:
            continue
        print >> fh,  (word_i,pos_i), v
    fh.close()

def read_ct_file (filename):
    """
    Parameter C{filename} is a frequency-sorted file (most frequent first)
    Return an item list preserving sort order.
    """
    fh = open(filename,'r')
    # il = [parse_ct_file_line(line,filename) for line in fh]
    il = []
    for line in fh:
        il.append(parse_ct_file_line(line,filename))
    fh.close()
    return il

## Allow float cts as in smoothed ct files.
number_re = r'\d*\.?\d+'
ct_file_line_re = re.compile(r'(\(.+\s*,\s*.+\))\s*(%s)' % (number_re,))
def parse_ct_file_line (line,filename,ct_type=int):
    line = line.strip()
    match = re.match(ct_file_line_re,line)
    if match:
        groups = match.groups()
        return (eval(groups[0]), ct_type(groups[1]))
    else:
        raise Exception, 'file %s: Ct file line %s in wrong format.' % (filename,line)

#    (pos_str, top_n_str) = dep_file_utils.get_file_suffixes(pos,  top_n)

def get_word_ranks (freq_file, wordset):
   import scipy.stats
   ct_il = read_ct_file(freq_file)
   rank_dict = {}
   for (i,(lex,ct)) in enumerate(ct_il):
       rank_dict[lex] = (i+1)

   rank_sum = 0
   rank_prod = 1
   harm_sum = 0
   sample_size = len(wordset)
   word_ranks = []
   for lex in wordset:
       try:
           rank_sum += rank_dict[lex]
           rank_prod *= rank_dict[lex]
           harm_sum += pow(rank_dict[lex],-1)
           word_ranks.append(rank_dict[lex])
       except KeyError:
           print 'Unknown word: %s' % (lex,)
   print 'Arith avg word rank in wordset: %.3f' % (float(rank_sum)/sample_size,)
   print 'Geom mean word rank in wordset: %.6f' % (pow(rank_prod, 1/(float(sample_size))),)
   print 'Harm mean word rank in wordset: %.6f' % (pow(harm_sum/(float(sample_size)),-1),)
   print 'Mean word rank in wordset: %.3f' % (scipy.stats.stats.nanmedian(np.array(word_ranks)),)
   return rank_dict
            
def get_all_ranks (word_deprel_ct_dict, pos='NN', newfile=None,top_n=None):
    """
    typical val for C{newfile}::
      -- os.path.join(swidsas_dir,'dep_data','dependency_DB_all_NN_freq.items')
    """
    il = sort_word_deprel_dict (word_deprel_ct_dict)
    if top_n:
        il = il[top_n:]
    if pos:
        il = [((lex,pos),ct) for ((lex,pos),ct) in il if pos == pos]
    if newfile:
        with open(newfile,'w') as ofh:
            for ((word,pos),v) in nil:
                print >> ofh, (word,pos), v
    return il

def open_and_make_backup(dd_file):
    """
    C{dd_file} is complete string path to a file.
    
    Open C{dd_file} but make a backup with an index in the name if
    dd_file exists. Preserve time of creation (Unix only!)

    Return open file handle for C{dd_file}.
    """
    if os.path.isfile(dd_file):
        copy2(dd_file,find_unique_backup_name(dd_file))
    return open(dd_file,'w')

def make_backup(dd_file):
    """
    C{dd_file} is complete string path to a file.
    
    Open C{dd_file} but make a backup with an index in the name if
    dd_file exists. Preserve time of creation (Unix only!)

    Return open file handle for C{dd_file}.
    """
    if os.path.isfile(dd_file):
        copy2(dd_file,find_unique_backup_name(dd_file))
    return dd_file


file_ctr_suffix_re = re.compile(r'(.*)_(\d+)$')
def find_unique_backup_name(full_fn):
    while os.path.isfile(full_fn):
        (base,ext) = os.path.splitext(full_fn)
        index = 0
        m = file_ctr_suffix_re.match(base)
        if m:
            base = m.groups()[0]
            index = int(m.groups()[1])
        full_fn = '%s_%d%s' % (base, index+1, ext)
    return full_fn

def get_file_suffixes(pos,  top_n):
    if pos:
        pos_str = '_%s' % pos
    else:
        pos_str = ''
    if top_n:
        if type(top_n) == tuple:
            top_n_str = '_%s_%s' % (top_n[0],top_n[1])
        else:
            top_n_str = '_%s' % top_n
    else:
        top_n_str = ''
    return (pos_str, top_n_str)


def freq_filter(lexemes,vocab_ctr,k):
    return [lex for lex in lexemes if vocab_ctr[lex] > k]

class Ctr (object):

    def __init__(self, init_val=0):
        self._val = init_val
        self._init_val = init_val

    def inc (self, inc_val=1):
        self._val += inc_val

    def get_next (self):
        self._val += 1
        return self._val

    def get_val (self):
        return self._val

    def set_val (self,new_val):
        self._val = new_val

    def reset (self):
        self._val = self._init_val
