"""
Key data structures built are rel_dict and decodebook.

decodebook[rel_index]

return the (word, deprel, pos) triple asociated with index.

rel_dict[(word,deprel,pos)]

returns the in dex for that triple.

C{get_deprel_info} builds both, given a dependency_db keys file.

"""

import dep_file_utils
from dep_file_utils import inverse_deprel_char as inv_char

# Omitted non hyphen cases P, EXT, SUFFIX, EXTR, MNR, PRN, OPRD, PRP, BNF, PUT, LGS
rel_strings = ['ADV','AMOD','LOC','NMOD','PMOD','PRD','OBJ','SBJ','PRT','TMP','COORD','DIR','SUB','DEP','VOC','DTV','APPO']

inverse_rels = set([])
inverse_rel_codes = set([])
all_rels = set([])
rel_dict = dict()
decodebook = dict()
rel_strings = set([])
decodebook = [('dummy_wd','dummy_dep_rel','dummy_pos')]

def make_dependency_db_codebooks(dependency_keys_file):
    """
    Use info in C{dependency_keys_file} to create a
    list of dependency_triples for decoding integer vals in dependency_db
    files.  A dependency triple is of the form (wd, deprel,pos).
    decode_bk[n] gives the the dependency_triple for the nth dep_rel.
    """
    global decodebook
    
    ## Start with a dummy val so that feature index = line number in feats file
    decodebook = [('dummy_wd','dummy_dep_rel','dummy_pos')]
    dep_rel_ctr = dep_file_utils.Ctr()
    db_keys_fh = open(dependency_keys_file,'r') 
    for line in db_keys_fh:
        dep_rel_ctr.inc()
        line = line.strip().split()
        (wd,dep_rel,pos) = dep_file_utils.parse_dep_rel_key_line(line,dep_rel_ctr.get_val())
        decodebook.append((wd,dep_rel,pos))
    return dep_rel_ctr.get_val()



def old_get_deprel_info (dependency_keys_file):
    """
    Generalized version of make_dependency_db_codebooks.
    Does more.  Stores data in the globals listed below.
    
    Typical value for C{dependency_keys_file}:
    
    filename = '../dep_data/dependency_DB.keys'
    """
    global inverse_rels, inverse_rel_codes, all_rels, rel_dict, decodebook, rel_strings,inv_char
    inverse_rels = set([])
    inverse_rel_codes = set([])
    all_rels = set([])
    #rel_fn = [ ]
    #for dr_str in rel_strings:
    #    rel_fn.append((dr_str,inv_char + dr_str,set([]),set([])))
    rel_dict = {}
    for  dr_str in rel_strings:
        rel_dict[dr_str] = set([])
        rel_dict[inv_char + dr_str] = set([])
    fh = open(dependency_keys_file, 'r')
    decodebook = [('dummy_wd','dummy_dep_rel','dummy_pos')]
    dep_rel_ctr = dep_file_utils.Ctr()
    for line in fh:
        dep_rel_ctr.inc()
        (wd,dep_rel,pos) = dep_file_utils.parse_dep_rel_key_line(line,dep_rel_ctr.get_val())
        decodebook.append((wd,dep_rel,pos))
        dep_rel_code = dep_rel_ctr.get_val()
        all_rels.add(dep_rel)
        if dep_rel.startswith(inv_char):
            inverse_rel_codes.add(dep_rel_code)
            inverse_rels.add(dep_rel)
        if dep_rel in rel_dict:
            rel_dict[dep_rel].add(dep_rel_code)
        #for (dr_str,inv_dr_str,dr_set,inv_dr_set) in rel_fn:
        #    if dep_rel == dr_str:
        #        dr_set.add(dep_rel_code)
        #        break
        #    elif dep_rel == inv_dr_str:
        #        inv_dr_set.add(dep_rel_code)
        #        break
    fh.close()

def get_deprel_info (dependency_keys_file):
    """
    Generalized version of make_dependency_db_codebooks.
    Does more.  Stores data in the globals listed below.
    
    Typical value for C{dependency_keys_file}:
    
    filename = '../dep_data/dependency_DB.keys'
    """
    global inverse_rels, inverse_rel_codes, all_rels, rel_dict, \
           decodebook, rel_strings,inv_char
    #rel_fn = [ ]
    #for dr_str in rel_strings:
    #    rel_fn.append((dr_str,inv_char + dr_str,set([]),set([])))
    # for  dr_str in rel_strings:
    #    rel_dict[dr_str] = set([])
    #    rel_dict[inv_char + dr_str] = set([])
    fh = open(dependency_keys_file, 'r')
    dep_rel_ctr = dep_file_utils.Ctr()
    for line in fh:
        dep_rel_ctr.inc()
        (wd,dep_rel,pos) = dep_file_utils.parse_dep_rel_key_line(line,dep_rel_ctr.get_val())
        decodebook.append((wd,dep_rel,pos))
        dep_rel_code = dep_rel_ctr.get_val()
        all_rels.add(dep_rel)
        if dep_rel.startswith(inv_char):
            inverse_rel_codes.add(dep_rel_code)
            inverse_rels.add(dep_rel)
        else:
            rel_strings.add(dep_rel)
    for dep_rel_code in range(len(decodebook)):
        (wd,dep_rel,pos) = decodebook[dep_rel_code]
        code_set = rel_dict.setdefault(dep_rel,set([]))
        code_set.add(dep_rel_code)
        #for (dr_str,inv_dr_str,dr_set,inv_dr_set) in rel_fn:
        #    if dep_rel == dr_str:
        #        dr_set.add(dep_rel_code)
        #        break
        #    elif dep_rel == inv_dr_str:
        #        inv_dr_set.add(dep_rel_code)
        #        break
    fh.close()
    return (rel_dict, decodebook)

def get_rel_set (dr):
    return rel_dict[dr][set_index]


def collect_mod_freq_info (deprel_dict):
    """
    Build cts for each lexeme recording token ct and modifier ct.
    
    Assumption: C{get_deprel_info} has been executed, generating the
    rel_dict.

    C{rel_strings} is a global list of rel names,
    """
    global inv_char, mod_freq_dict
    mod_freq_dict = {}
    for lexeme in  deprel_dict:
        #print lexeme, deprel_dict[lexeme]
        token_ctr = 0
        mod_ctr = 0
        mdict = { }
        mod_freq_dict[lexeme] = mdict
        this_dict = deprel_dict[lexeme]
        for (deprel_code,deprel_ct) in  this_dict.iteritems():
            #print lexeme, deprel_code, type(deprel_code)
            deprel = decodebook[deprel_code][1]
            if deprel in mdict:
                mdict[deprel] += deprel_ct
            else:
                mdict[deprel] = deprel_ct
            if deprel.startswith(inv_char):
                token_ctr += deprel_ct
            else:
                mod_ctr += deprel_ct
        mdict['token_ctr'] = token_ctr
        mdict['mod_ctr'] = mod_ctr
    return mod_freq_dict

def sort_deprel_freqs (deprel_dict,mod_freq_dict=None,top_n=3000):
    """
    Build cts for each lexeme recording token ct and modifier ct.

    Assumption 1: mod_freq_dict passed in or computable.
    
    Assumption 2: Computabliliti of mod_freq_dict presupposes there is a C{rel_dict}
    globally available. Which means C{get_deprel_info} has been executed, generating the
    C{rel_dict}.
    """
    global mod_freq_dict1
    if mod_freq_dict is None:
        print "Collecting mod freq info"
        mod_freq_dict = collect_mod_freq_info (deprel_dict)
        print 'mod_freq info collected'
    mod_freq_dict1 = mod_freq_dict
    token_frequencies = mod_freq_dict.items()
    mod_frequencies = mod_freq_dict.items()
    print 'Getting token frequencies'
    token_frequencies.sort(key=lambda x: x[1]['token_ctr'],reverse=True)
    print 'Getting mod frequencies'
    mod_frequencies.sort(key=lambda x: x[1]['mod_ctr'],reverse=True)
    (v_ctr,n_ctr) = (1,1)
    for ((f,pos),mdict) in token_frequencies:
        if v_ctr > 10 and n_ctr > 10:
            break
        if pos == 'NN':
            print 'The %dth most frequent noun is %s %s' % (n_ctr, f,pos)
            n_ctr += 1
        elif pos.startswith('V'):
            print 'The %dth most frequent verb form is %s %s' % (v_ctr, f,pos)
            v_ctr += 1
    (v_ctr,n_ctr) = (1,1)
    best_ratio_n,best_ratio_v,n_total,v_total = 1.0,1.0,0.0,0.0
    best_n, best_v = (None,None)
    mod_pos_ctr,freq_pos_ctr, pos_ctr = {},{},{}
    for ((f,pos),mdict) in mod_frequencies[:top_n]:
        mod_ct, token_ct = mdict['mod_ctr'], mdict['token_ctr']
        if pos in mod_pos_ctr:
            mod_pos_ctr[pos] += mod_ct
            freq_pos_ctr[pos] += token_ct
            pos_ctr[pos] += 1
        else:
            mod_pos_ctr[pos] = mod_ct
            freq_pos_ctr[pos] = token_ct
            pos_ctr[pos] = 1
        ratio = float(mod_ct)/token_ct
        if pos == 'NN':
            # print 'The %dth most frequently modified  noun is %s %s' % (n_ctr, f,pos)
            n_ctr += 1
            n_total += ratio
            if ratio > best_ratio_n:
                best_ratio_n = ratio
                best_n = (f,pos)
        elif pos.startswith('V'):
            # print 'The %dth most frequently modified verb form is %s %s' % (v_ctr, f,pos)
            v_ctr += 1
            v_total += ratio
            if ratio > best_ratio_v:
                best_ratio_v = ratio
                best_v = (f,pos)
    return (token_frequencies,mod_frequencies,\
            best_n,best_ratio_n,\
            best_v,best_ratio_v,\
            n_total/n_ctr,\
            v_total/v_ctr,\
            pos_ctr,mod_pos_ctr,freq_pos_ctr)
    
    

def print_deprel_profile (lexeme,deprel_dict,ct_dict=True,  print_mod_pos=False,pairs=False):
    """
    Assumption: C{} has been executed, generating the
    rel_dict.

    C{rel_strings} is a global list of rel names,
    """
    global rel_strings
    word_vec = deprel_dict[lexeme]
    print_deprel_profile_vec (word_vec,ct_dict=ct_dict, \
                              print_mod_pos=print_mod_pos,pairs=pairs)

def print_deprel_profile_vec (word_vec,ct_dict=True,  print_mod_pos=False,pairs=False):
    for deprel in rel_strings:
        print_word_deprels(deprel,word_vec,ct_dict=ct_dict,print_mod_pos=print_mod_pos,pairs=pairs)
        print_word_deprels(inv_char+deprel,word_vec,ct_dict=ct_dict,print_mod_pos=print_mod_pos,pairs=pairs)


def compare_word_vectors (lex1,lex2, deprel_dict,ct_dict=False,print_mod_pos=False):
    lex1_vec = deprel_dict[lex1]
    lex2_vec = deprel_dict[lex2]
    merged = dict()
    for (feat, val1) in lex1_vec.iteritems():
        if feat in lex2_vec:
            merged[feat] = (val1,lex2_vec[feat])
    print 'Lex1: %d Lex2: %d Shared: %d' % (len(lex1_vec),len(lex2_vec),len(merged))
    print_deprel_profile_vec(merged,ct_dict=ct_dict,print_mod_pos=print_mod_pos,pairs=True)
    
def get_relatum(rel_code,print_mod_pos):
    global rel_dict, decodebook

    if print_mod_pos:
        ## Get pos as well. Make lower case.
        return '%s.%s' % (decodebook[rel_code][0],decodebook[rel_code][2].lower())
    else:
        return decodebook[rel_code][0]

def print_word_deprels(deprel,word_vec,ct_dict=True,print_mod_pos=False,pairs=False):
    """
    Changed so the values of a feat can be a pair of floats, signaled by C{pairs} = True.
    Useful when printing comparison showing shared feats of two lexemes.
    """


    global rel_dict, decodebook

    if pairs:
        if ct_dict:
            pair_format_str = '%s:(%d,%d)'
        else:
            pair_format_str = '%s:(%.4f,%.4f)'
    else:
        if ct_dict:
            pair_format_str = '%s:%d'
        else:
            pair_format_str = '%s:%.4f'
    try:
        rel_set = rel_dict[deprel]
    except NameError,e:
        print '**** You must execute get_deprel_info before print_word_deprel will work. ****'
        raise NameError, e
        
    sorted_ilist = word_vec.items()
    if pairs:
        sorted_ilist.sort(key = lambda x:x[1][0],reverse=True)
        str_list = [pair_format_str % (get_relatum(rel_code,print_mod_pos),rel_ct[0],rel_ct[1]) \
                    for (rel_code,rel_ct) in sorted_ilist if rel_code in rel_set]
    else:
        sorted_ilist.sort(key = lambda x:x[1],reverse=True)
        str_list = [pair_format_str % (get_relatum(rel_code,print_mod_pos),rel_ct) \
                    for (rel_code,rel_ct) in sorted_ilist if rel_code in rel_set]
    if str_list:
        intro_str = '%s: ' % deprel
        max_ind = len(str_list)
        inc_len = len(intro_str)
        indent_str = '%*s' % (inc_len,'')
        index = 0
        total_len = inc_len
        print intro_str,
        while index < max_ind:
            while total_len < 60 and index < max_ind:
                cur_chunk = str_list[index]
                print cur_chunk,
                total_len += len(cur_chunk)
                index += 1
            print
            print indent_str,
            total_len = 0
        print

def make_relations_key (dependency_keys_file):
    decodebook = [('dummy_wd','dummy_dep_rel','dummy_pos')]
    relations_key = {}
    dep_rel_ctr = dep_file_utils.Ctr()
    db_keys_fh = open(dependency_keys_file,'r') 
    for line in db_keys_fh:
        dep_rel_ctr.inc()
        line = line.strip()
        (wd,dep_rel,pos) = dep_file_utils.parse_dep_rel_key_line(line,dep_rel_ctr.get_val())
        decodebook.append((wd,dep_rel,pos))
        relations_key[(wd,dep_rel,pos)] =  dep_rel_ctr.get_val()
    return relations_key


def get_triple_for_code (deprel):
    (wd,dep_rel,pos) = decodebook[deprel]
    return (wd,dep_rel,pos)


if __name__ == '__main__':
    home = os.environ['HOME']
    dependency_db_dir = os.path.join(home,'Desktop','SWIDSAS','dependency_DB')
    os.chdir(dependency_db_dir)

    ## Assumed to exist
    dependency_keys_file = '../dep_data/dependency_DB.keys'
    rk = make_relations_key (dependency_keys_file)
