import os.path, sys,math,numpy
from dependency_db import read_dependency_file, reset_dep_db_globals
from dep_file_utils import sort_word_deprel_dict
from dep_rel_info import get_deprel_info, print_deprel_profile, print_deprel_profile_vec

def make_rank_dict (vocab):
    """
    Pass in a sorted vocab of word count pairs and turn
    it in to a dictionary assigning a rank to each word.
    """
    print 'Constructing rank dict'
    rank_dict = {}
    for (i,(lex,ct)) in enumerate(vocab):
        rank_dict[lex] = i+1
    print 'Rank & Count dicts constructed'
    return rank_dict


def show_shared_properties (lex1, lex2, deprel_dict):
    vec1 = deprel_dict[lex1]
    vec2 = deprel_dict[lex2]
    sum_sqs = sum((v**2 for v in vec1.values()))
    S = set(vec1.keys()).intersection(set(vec2.keys()))
    print '%d shared features!: ss: %.5f dp: %.5f dm: %.5f' % (len(S),sum_sqs,sim_measures.dot_prod(vec1,vec2),sim_measures.dot_min(vec1,vec2))
    print
    fw = {}
    for f in S:
        fw[f] = (vec1[f],vec2[f])
    dummy_lex = ('dummy_word','NN')
    deprel_dict[dummy_lex] = fw
    print_deprel_profile(dummy_lex,deprel_dict,ct_dict=False,pairs=True)
    del deprel_dict[dummy_lex]


def modify_vector (vec,concentration_dict,decodebook):
    new_vec = {}
    for (k,v) in vec.items():
        (wd,dep_rel,pos) = decodebook[k]
        new_vec[k] = math.log(concentration_dict[(wd,pos)],2) + v
    return new_vec


def bnc_n (pos_in):
    #poses =  pos_in.split('/')
    #for pos in poses:
    if  pos.startswith('NN1'):
        return True
    elif pos == 'PNI':
        return True
    #elif pos == 'NP0':
    #    return True
    else:
        return False
    
       

def bnc_n2penn_n(pos,pred=bnc_n):
    if bnc_n(pos):
        return 'NN'
    else:
        return pos




def vals_to_pdist (vals):
    tot = float(sum(vals))
    return [v/tot for v in vals]

def entropy (pdist):
    return sum((p*info_val(p) for p in pdist))


def info_val (p):
    return -math.log(p,2)

def vals_to_entropy (vals):
    return entropy(vals_to_pdist(vals))

def scaled_entropy (vals):
    max_pt = math.log(len(vals),2)
    pdist = vals_to_pdist(vals)
    try:
        scaled_ent = entropy(pdist)/max_pt
    except ZeroDivisionError:
        return 1.0
    return scaled_ent


def avg (vals):
    return float(sum(vals))/len(vals)


def print_ratio_dict():
    with open('dep_ratio.dat','w') as fh:
        ctr = 0
        err_ctr = 0
        anomaly_ctr = 0
        for (lex,ct) in vocab:
            ctr +=1
            if token_ct_dict[lex] > count_dict[lex]:
                anomaly_ctr += 1
                print 'Count anomaly: %s raw count: %d dep count %d' % (lex,revised_token_ct_dict[lex], count_dict[lex])
                find_homonym(lex[0],initial_token_ct_dict)
            try:
                print >> fh, '%-8.5f  %.5f' % (math.log(rank_dict[lex]),math.log(float(count_dict[lex])/float(revised_token_ct_dict[lex])))

            except KeyError,e:
                err_ctr += 1
                print '%s not in  revised_token_ct_dict' % (lex,)
                find_homonym(lex[0], token_ct_dict)
                if err_ctr > 300:
                    raise e
        print '%d out of %d errors [%d anomalies]' % (err_ctr, ctr, anomaly_ctr)


def count_corpus (filename,pos_map=bnc_n2penn_n,pos_pred=bnc_n):
    token_ct_dict = {}
    with open(filename,'r') as fh:
        for line in fh:
            line = line.strip()
            for wt in line.split():
                try:
                    (word,tag) = wt.split('_')
                except:
                    continue
                word = word.lower()
                if (word,tag) in token_ct_dict:
                    token_ct_dict[(word,tag)] += 1
                else:
                    token_ct_dict[(word,tag)] = 1
    il = token_ct_dict.items()
    il.sort(reverse=True, key = lambda x: x[1])
    il = [((lex,pos_map(pos)),ct) for ((lex,pos),ct) in il if pos_pred(pos)]    

    token_rank_dict = {}
    for (i,(lex,ct)) in enumerate(il):
        token_rank_dict[lex] = (i + 1)
    
    il2 = token_ct_dict.items()
    revised_token_ct_dict = {}
    for ((wd,pos),ct) in il2:
        revised_token_ct_dict[(wd,pos_map(pos))] = ct
    return (revised_token_ct_dict,token_ct_dict,token_rank_dict,il)


def find_homonym (wd, wd_dict):
    for (wd0, pos0) in wd_dict:
        if wd0 == wd:
            print '%-25s %5s %d' % (wd0, pos0, wd_dict[(wd0,pos0)])
            



if __name__ == '__main__':

    home = os.getenv('HOME')
    dependency_db_dir = os.path.join(home,'Desktop','SWIDSAS','dependency_DB')
    os.chdir(dependency_db_dir)

    swidsas_dir = os.path.join(home,'Desktop','SWIDSAS')
    output_dir = os.path.join(swidsas_dir,'dependency_db_test_results','new_bulba_nn_files','0_10000')
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    external_nn_dir = os.path.join(swidsas_dir,'dependency_db_test_results','new_bulba_nn_files','0_5000')

    dependency_keys_file = '../dep_data/dependency_DB.keys'
    dependency_train_file = '../dep_data/dependency_DB.train'


    #vals0 = [25,25,25,25]
    #vals1= [13,25,37,25]
    #vals2 = [2,8,3,7,5,15,15,15,15,15]
    #vals3 = [10]*10
    #vals4 =[100]
    #vals5 = range(1,10)
    #vals6 = [3*x for x in vals5]

    #for (i,l) in enumerate([vals0,vals1,vals2,vals3,vals4,vals5,vals6]):
    #    print 'vals%d  %9.5f %3.5f' % (i, entropy(l), avg(l))

    
    
    print 'Reading in a count model'
    (deprel_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
                          read_dependency_file(dependency_train_file, count_file=True)

    print 'Sorting vocab by counts'
    raw_vocab = sort_word_deprel_dict (deprel_dict)
    print 'Vocab sorted'

    pos = 'NN'
    top_n = 10000

    #vocab = [((wd,pos2),ct) for ((wd,pos2),ct) in raw_vocab if pos2 == pos]
    vocab = raw_vocab

    print 'Constructing rank dict'
    rank_dict = make_rank_dict(vocab)
    count_dict = dict(vocab)
    #rank_dict = {}
    print 'Rank & Count dicts constructed'

    avg_dep_ct_dict = dict()
    concentration_dict = dict()
    for (lex, ct) in vocab:
        promisc = len(deprel_dict[lex])
        cnctrn = float(ct)/promisc
        adj_cnctrn = cnctrn/(1 + math.log(rank_dict[lex] + 1))
        #adj_cnctrn = cnctrn/(promisc)
        #stdev = numpy.std(deprel_dict[lex].values())
        #ent = scaled_entropy(deprel_dict[lex].values())
        avg_dep_ct_dict[lex] = (rank_dict[lex],cnctrn,adj_cnctrn,promisc)
        concentration_dict[lex] = cnctrn


    reset_dep_db_globals()

    ############################################
    ##    S e t  w e i g h t  f u n c t i o n
    ############################################
    #wt_func = 't_test_no_norm'
    #wt_func = 't_test'

    wt_func = 'pmi'
    #wt_func = 'pmi_no_norm'
    #wt_func = 'pmi_poiss_no_norm'
    #wt_func = 'pmi_poiss'
    #wt_func = 'qw_poiss'
    #wt_func = 'lpmi'
    wt_func = 'pmi_no_norm'

    if wt_func:
        wt_func = '_' + wt_func

    ############################################
    ##    S e t   s u f f i x  (s i m f n)
    ############################################
        
    #suffix = 'dot_prod'
    #suffix = 'cos'
    suffix = ''

    if suffix:
        suffix = '_' + suffix

    print 'dependency train file: %s' % (dependency_train_file,)
    sqs_dict = { }
    #print 'Reading in squares dict for model %s' % (wt_func,)
    #sqs_dict = read_sqs_file(dependency_train_file, suffix = '_' + suffix, ext = '_' + wt_func)
    #print 'Squares dict read in'

    

    print 'Reading in a %s  model' % (wt_func,)
    #suffix='',ext='_qw_poiss'
    (deprel_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
                  read_dependency_file(dependency_train_file, count_file=False, \
                                       suffix = suffix, ext = wt_func)
    print 'Model read in'

    get_deprel_info(dependency_keys_file)
    from dep_rel_info import decodebook


    # Need to execute get_deprel_info to print a lex profile
    # must attend to ct_dict flag
    # print_deprel_profile(('papa','NN'), deprel_dict, ct_dict=False)

    # Need to execute get_deprel_info to print shared info
     #show_shared_properties(('rebuttal','NN'),('refutation','NN'),deprel_dict)
 
    vec = deprel_dict[('aspersions','NNS')]
    print_deprel_profile_vec(vec, ct_dict=False)

    new_vec = modify_vector (vec,concentration_dict,decodebook)

    print_deprel_profile_vec(new_vec, ct_dict=False)

    #il = avg_dep_ct_dict.items()
    #il.sort(key=lambda x:x[1][1])
    #winners = il[-300:]
    
    #oil = avg_dep_ct_dict.items()
    #oil.sort(key=lambda x:x[1][2],reverse=True)
    #other_winners = oil[:300]
    #other_winners = oil[:300]
    
    # con = concentration [antonym of dispersion] avg number of token/dependent
    # promisc.  Number of dependent types.
    #winners = [(l,info) for (l,info) in winners if count_dict[l] >= 50]

    #for w in winners:
    #    print '%-20s %4d %4d' % (w[0][0], w[1][0], w[1][3])


    #print
    #print

    #for w in other_winners:
    #    print '%-20s %4d %4d' % (w[0][0], w[1][0], w[1][3])        

    #bnc_texts = os.path.join(home,'ext','corpus','bnc','written.tag')
    #bnc_texts = os.path.join(home,'ext','corpus','bnc','short_written.tag')
    #(token_ct_dict,initial_token_ct_dict,token_rank_dict,token_vocab) = count_corpus(bnc_texts)
    #print_ratio_dict()
    #raise Exception

