# -*- coding: utf-8 -*-
from dependency_db import  read_dependency_file, reset_dep_db_globals, sim_score,read_nn_file_to_list_safe
from dep_rel_info import get_deprel_info, print_deprel_profile, print_deprel_profile_vec
from dep_file_utils import sort_word_deprel_dict
import sim_measures
import math, time, os.path
from datetime import datetime
from get_dependency_stats import make_rank_dict
from collections import Counter



def sim_score_words (w1,w2, deprel_dict, sim_func, pos='NN'):
    """
    sim_func is string name
    """
    return sim_score((w1,pos),(w2,pos),sim_func=sim_func,deprel_dict=deprel_dict)



def update_nn_dict_from_file(nn_file,nn_dict):
    if os.path.isfile(nn_file):
        try:
            read_nn_file_safe(nn_file,nn_dict=nn_dict)
        except NNError, e:
            if os.path.getsize(nn_file) <> 0:
                raise e    

def relative_best_nn (lex, lexset, deprel_dict, sim_func):
    best = ('None','NoPOS')
    best_score = 0
    print len(lexset)
    for (lex2,pos2) in lexset:
        if lex2 == lex:
            continue
        this_score = sim_score_words(lex,lex2,deprel_dict,sim_func)
        if this_score > best_score:
            best_score = this_score
            best = lex2
    return (best,best_score)

def get_nn_list (sys_name, nn_dir=''):
    return read_nn_file_to_list_safe(\
        os.path.join(nn_dir,'nearest_neighbors_10000_%s.dict' % sys_name))

def get_nn_stats (nn_list,rank_dict):
    rank_tot = 0
    ctr = 0
    nn_ctr = Counter()
    for (lex, (nn, score)) in nn_list:
        ctr += 1
        rank_tot += rank_dict[nn]
        nn_ctr[nn] += 1
    return (float(rank_tot)/ctr,nn_ctr)


def graph_avg_nn (nn_list, rank_dict, data_file=None, incs=100):
    """
    Because of key errors in the rank dict on unknown
    nns, we may not actually have C{inc} NEW
    data points when C{i%incs == 0}, but we were still
    averaging over the last C{inc} datapoints
    to get our bin number, possibly overlapping slightly with
    the last bin.

    Alternative now implemented is to keep a separate ctr tracking valid
    data points and bin as follows::
     if i%incs == 0:
            bins.append((i, float(sum(data[-ctr:]))/ctr))
            ctr = 0
    Lots more complicated.
    """
    data = []
    bins = []
    ctr = 0 
    for (i,(t, (nn, score))) in enumerate(nn_list):
        if i%incs == 0 and ctr > 0:
            bins.append((i, float(sum(data[-ctr:]))/ctr))
            ctr = 0
        try:
            data.append(rank_dict[nn])
        except KeyError:
            continue
        ctr += 1
    if data_file:
        with open(data_file, 'w') as ofh:
            for (t_bin, avg_nn_rank) in bins:
                print >> ofh, '%d  %.1f' % (t_bin, avg_nn_rank)
    return (bins, data)


def find_all_nns (nn_list, lex):
    ctr = 0
    res = []
    for (t,(nn,s)) in nn_list:
        if nn == lex:
            ctr += 1
            res.append(t)
            print '%4d. %s' % (ctr,t)
    return res


def compare_nn_lists (nn_list1,nn_list2):
    for (i, (t, (nn, score))) in enumerate(nn_list1):
        nn2 = nn_list2[i][1][0]
        print '%4d.  %24s  %24s[%4d] %24s[%4d]' % (rank_dict[t], t[0], nn[0], rank_dict[nn],nn2[0],rank_dict[nn2])

if __name__ == '__main__':

    home = os.getenv('HOME')

    root_dir = os.path.join(home,'Desktop','SWIDSAS')
    dependency_db_dir = os.path.join(root_dir, 'dependency_DB')
    os.chdir(dependency_db_dir)
    dep_db_dir = '../dep_data'
    #dep_db_dir = '../dep_data_extra'
    #dep_db_dir = os.path.join(home,'ext/corpus/bnc_stanford_dep_db/')
    #dep_db_dir = os.path.join(home,'Desktop/SWIDSAS/dep_data_extra/')


    #output_dir = os.path.join(root_dir,'dependency_db_test_results','new_bulba_nn_files','0_10000')
    output_dir = os.path.join(root_dir,'dependency_db_test_results','new_bulba_nn_files','final_10000')
    #output_dir = os.path.join(root_dir,'dependency_db_test_results','new_bulba_nn_files','0_10000_test')
    #output_dir = os.path.join(root_dir,'dependency_db_test_results','new_bulba_nn_files', '0_10000_rank_dict')
    
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    parse_domain = 'dependency_DB'
    #parse_domain = 'bnc_stanford'
    #parse_domain= 'bnc_xml_DB'
    
    #extra= '_pruned'
    #extra= '_pruned_nonneg'
    extra = ''

    dependency_keys_file = os.path.join(dep_db_dir,'%s.keys' % (parse_domain,))
    dependency_train_file = os.path.join(dep_db_dir, '%s%s.train' % (parse_domain,extra))

    print 'dependency train file: %s' % (dependency_train_file,)
    
    print 'Reading in a count model'
    (deprel_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
                          read_dependency_file(dependency_train_file, count_file=True)
    def weird_word (wd):
        return wd in """][`'-_?.,;!~*()/\)"""


    print 'Sorting vocab by counts'
    raw_vocab = sort_word_deprel_dict (deprel_dict)
    print 'Vocab sorted'

    pos = 'NN'
    top_n = 10000
    #top_n = 500

    vocab = [((lex,pos2),ct) for ((lex,pos2),ct) in raw_vocab \
                     if ((pos2 == pos) and not weird_word(lex))]

    #get_related_words_vocab(vocab[:top_n],'n',add_dfns=True,res_file=os.path.join(output_dir,'wordnet_related_words.txt'),track_progress=True)
    # Once youve computed the related words, just do this
    #wordnet_related_words_file = os.path.join(output_dir,'wordnet_related_words.txt')
    
    #raise Exception
    
    #print 'Constructing rank dict'
    rank_dict = make_rank_dict(vocab)
    #print 'Rank & Count dicts constructed'
    #rank_dict = {}

    ## Why? JMG
    count_dict = dict(vocab)


    ############################################
    ##    S e t  w e i g h t  f u n c t i o n
    ############################################
    #wt_func = 't_test_no_norm'
    #wt_func = 't_test'

    #wt_func = 'pmi'
    #wt_func = 'pmi_nonneg'
    #wt_func = 'pmi_no_norm'
    #wt_func = 'pmi_poiss_no_norm'
    #wt_func = 'pmi_poiss'
    #wt_func = 'qw_poiss'
    #wt_func = 'lpmi'
    #wt_func = 'hd_pmi_no_norm'
    wt_func = None

    if wt_func:
        wt_func = '_' + wt_func

    ############################################
    ##    S e t   s u f f i x  (s i m f n)
    ############################################
        
    #suffix = 'dot_prod'
    suffix = 'cos'
    #suffix = ''

    if suffix:
        suffix = '_' + suffix


    if wt_func:
    
        print 'Reading in a %s  model' % (wt_func,)
        #suffix='',ext='_qw_poiss'
        reset_dep_db_globals()
        
        (deprel_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
                      read_dependency_file(dependency_train_file, count_file=False, \
                                           suffix = suffix, ext = wt_func)
        print 'Wt Model read in'

        def weird_word (wd):
            return wd in """][`'-_?.,;!~*()/\)"""

        for (wd, pos2) in deprel_dict.keys():
            if weird_word(wd):
                del deprel_dict[(wd,pos2)]



    ####################################################################################################
    ####################################################################################################
    ##
    ##   S i m    F u n c t i o n s
    ##
    ####################################################################################################
    ####################################################################################################

    #sim_func = sim_measures.dice_dot_prod
    #sim_func = sim_measures.dice_dagger 
    #sim_func = sim_measures.dice_dot_prod_sq_rt
    #sim_func = sim_measures.dot_prod
    #sim_func = sim_measures.dot_prod_sq_rt
    #sim_func = sim_measures.dice_dot_prod_sq_rt
    sim_func = sim_measures.dot_min  
    #sim_func = sim_measures.dot_sum  
    #sim_func = sim_measures.lin_no_neg
    #sim_func = sim_measures.dot_min_length 

    ####################################################################################################
    ####################################################################################################
    ##
    ##   M i s c e l l a n e o u s    P a r a m s
    ##
    ####################################################################################################
    ####################################################################################################

    inspect_nn_files = True

    if inspect_nn_files:
        # A couple of base systems to play with 
        dot_min_no_norm = get_nn_list ('dot_min_pmi_no_norm', nn_dir=output_dir)
        dot_min_no_norm_dict = dict(dot_min_no_norm)
        dot_min = get_nn_list ('dot_min_pmi', nn_dir=output_dir)
        dot_min_dict = dict(dot_min)

        
        (dm_no_norm_avg_rank,dm_no_norm_nn_ctr) = get_nn_stats (dot_min_no_norm,rank_dict)
        #(dm_no_norm_avg_rank_50,dm_no_norm_nn_ctr_50) = get_nn_stats (dot_min_no_norm[-50:],rank_dict)


        (dm_avg_rank,dm_nn_ctr) = get_nn_stats (dot_min,rank_dict)
        #(dm_avg_rank_50,dm_nn_ctr_50) = get_nn_stats (dot_min[-50:],rank_dict)

        top_dot_min_no_norm_nns = dm_no_norm_nn_ctr.most_common(20)
        top_dot_min_nns = dm_nn_ctr.most_common(20)
        man_words = find_all_nns(dot_min_no_norm,('man','NN'))
        pauper_nn = relative_best_nn ('pauper', man_words, deprel_dict, 'dot_min')
        beggar_nn = relative_best_nn ('beggar', man_words, deprel_dict, 'dot_min')
        gnuplot_avg_rank_dir = os.path.join(output_dir, 'gnuplot_avg_rank')
        if not os.path.isdir(gnuplot_avg_rank_dir):
            os.mkdir(gnuplot_avg_rank_dir)
        nn_rank_template = '%s_avg_nn_rank.dat'

        ### Some other systems for avg nn rank
        UKB = get_nn_list ('dot_prod__ukb', nn_dir=output_dir)
        dot_prod = get_nn_list ('dot_prod_pmi_no_norm', nn_dir=output_dir)

        dice_dot_prod = get_nn_list ('dice_dot_prod_pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_alpha_30 = get_nn_list ('dice_dot_prod_alpha_30__pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_alpha_15 = get_nn_list ('dice_dot_prod_alpha_15__pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_alpha_07 = get_nn_list ('dice_dot_prod_alpha_07__pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_alpha_var = get_nn_list ('dice_dot_prod_alpha_var1500__pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_alpha_30symm = get_nn_list ('dice_dot_prod_alpha_30symm__pmi_no_norm', nn_dir=output_dir)
        dice_dot_prod_sq_rt = get_nn_list ('dice_dot_prod_sq_rt_pmi_no_norm', nn_dir=output_dir)

        dot_prod_sq_rt = get_nn_list ('dot_prod_sq_rt_pmi', nn_dir=output_dir)
        dot_prod_sq_rt_no_norm = get_nn_list ('dot_prod_sq_rt_pmi_no_norm', nn_dir=output_dir)

        dot_sum = get_nn_list ('dot_sum_pmi', nn_dir=output_dir)
        dot_sum_no_norm = get_nn_list ('dot_sum_pmi_no_norm', nn_dir=output_dir)

        dice_dagger = get_nn_list ('dice_dagger_pmi_no_norm', nn_dir=output_dir)
        lin_no_neg = get_nn_list ('lin_no_neg_pmi_no_norm', nn_dir=output_dir)

        gnuplot_ukb_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'ukb')
        
        gnuplot_dot_min_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_min')
        gnuplot_dot_min_no_norm_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_min_no_norm')
        
        gnuplot_dot_prod_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_prod')
        gnuplot_dice_dot_prod_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod')
        gnuplot_dice_dot_prod_alpha_30_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_alpha_30')
        gnuplot_dice_dot_prod_alpha_15_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_alpha_15')
        gnuplot_dice_dot_prod_alpha_07_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_alpha_07')
        gnuplot_dice_dot_prod_alpha_30symm_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_alpha_30symm')
        gnuplot_dice_dot_prod_alpha_var_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_alpha_var')
        gnuplot_dice_dot_prod_sq_rt_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dot_prod_sq_rt')
        gnuplot_dice_dagger_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dice_dagger')
        gnuplot_lin_no_neg_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'lin_no_neg')
        gnuplot_dot_prod_sq_rt_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_prod_sq_rt')
        gnuplot_dot_prod_sq_rt_no_norm_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_prod_sq_rt_no_norm')
        
        gnuplot_dot_sum_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_sum')
        gnuplot_dot_sum_no_norm_nn_rank_file = \
              os.path.join(gnuplot_avg_rank_dir, nn_rank_template % 'dot_sum_no_norm')
        

        graph_avg_nn (UKB, rank_dict, \
                      data_file=gnuplot_ukb_nn_rank_file)
        graph_avg_nn (dot_min, rank_dict, \
                      data_file=gnuplot_dot_min_nn_rank_file)
        graph_avg_nn (dot_min_no_norm, rank_dict, \
                      data_file=gnuplot_dot_min_no_norm_nn_rank_file)
        graph_avg_nn (dot_prod, rank_dict, \
                      data_file=gnuplot_dot_prod_nn_rank_file)
        graph_avg_nn (dice_dot_prod, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_nn_rank_file)
        graph_avg_nn (dice_dot_prod_alpha_30, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_alpha_30_nn_rank_file)
        graph_avg_nn (dice_dot_prod_alpha_15, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_alpha_15_nn_rank_file)
        graph_avg_nn (dice_dot_prod_alpha_07, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_alpha_07_nn_rank_file)
        graph_avg_nn (dice_dot_prod_alpha_30symm, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_alpha_30symm_nn_rank_file)
        graph_avg_nn (dice_dot_prod_alpha_var, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_alpha_var_nn_rank_file)
        graph_avg_nn (dice_dot_prod_sq_rt, rank_dict, \
                      data_file=gnuplot_dice_dot_prod_sq_rt_nn_rank_file)
        graph_avg_nn (dice_dagger, rank_dict, \
                      data_file=gnuplot_dice_dagger_nn_rank_file)
        graph_avg_nn (lin_no_neg, rank_dict, \
                      data_file=gnuplot_lin_no_neg_nn_rank_file)
        graph_avg_nn (dot_prod_sq_rt, rank_dict, \
                      data_file=gnuplot_dot_prod_sq_rt_nn_rank_file)
        graph_avg_nn (dot_prod_sq_rt_no_norm, rank_dict, \
                      data_file=gnuplot_dot_prod_sq_rt_no_norm_nn_rank_file)
        graph_avg_nn (dot_sum, rank_dict, \
                      data_file=gnuplot_dot_sum_nn_rank_file)
        graph_avg_nn (dot_sum_no_norm, rank_dict, \
                      data_file=gnuplot_dot_sum_no_norm_nn_rank_file)

