import math,sys
from collections import Counter

neg_infinity=float("-infinity")

####################################################################################
####################################################################################
##
##     Weight functions  (v1[fi] => v1[fi_weighted])
##
####################################################################################
####################################################################################

def counts_to_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                           non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    4 counts needed.  All are assumed to be of disjoint sets, the counts
    of the T,C, T, not C, not T C, and not T not C.  Since the 4 possible
    outcomes for random variables T and C are exhausted, the sum of all 4 counts is
    the total number of events, called N.
    """
    ## The two row counts
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct
    ## The two col counts
    term_ct = class_term_ct + non_class_term_ct
    non_term_ct = class_non_term_ct + non_class_non_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    val_list = [N, class_ct, non_class_ct, term_ct, non_term_ct, class_term_ct, \
                class_non_term_ct,non_class_term_ct, non_class_non_term_ct]
    ## test for bug:::
    if debug:
        for val in val_list:
            if val < 0:
                raise Exception, lexeme, val_list
    ## A true constraint
    ## assert N = term_ct + non_term_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N.0
    non_class_prob = float(non_class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N0.
    nonterm_prob = float(non_term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    # N10
    non_class_term_joint_prob = float(non_class_term_ct)/N
    # N01
    class_nonterm_joint_prob = float(class_non_term_ct)/N
    # N00
    non_class_nonterm_joint_prob = float(non_class_non_term_ct)/N
    ## Weighted sums
    pos_class_sum = (class_term_joint_prob * \
                     my_log(class_term_joint_prob/(term_prob * class_prob))) + \
                    (class_nonterm_joint_prob * \
                     my_log(class_nonterm_joint_prob/(nonterm_prob * class_prob))) 
    neg_class_sum = (non_class_term_joint_prob * \
                     my_log(non_class_term_joint_prob/(term_prob * non_class_prob))) + \
                    (non_class_nonterm_joint_prob * \
                     my_log(non_class_nonterm_joint_prob/(nonterm_prob * non_class_prob)))
    res = pos_class_sum + neg_class_sum
    #  print 'class_term_joint_prob: %.7f (%.7f)' % (class_term_joint_prob,float(class_term_ct)/class_ct)
    # print 'non_class_term_joint_prob: %.7f (%.7f)' % (non_class_term_joint_prob, float(non_class_term_ct)/non_class_ct)
    # print 'MI: %.7f' % res 
    return res

def counts_to_info (lexeme, class_term_ct, class_non_term_ct, \
                           non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    4 counts needed.  All are assumed to be of disjoint sets, the counts
    of the T,C, T, not C, not T C, and not T not C.  Since the 4 possible
    outcomes for random variables T and C are exhausted, the sum of all 4 counts is
    the total number of events, called N.
    """
    ## The two row counts
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct
    ## The two col counts
    term_ct = class_term_ct + non_class_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    return my_log ((float(class_term_ct)/term_ct)/(float(class_ct)/N))

def counts_to_pointwise_mutual_info_sqrt (lexeme, class_term_ct, class_non_term_ct, \
                                          non_class_term_ct, non_class_non_term_ct, \
                                          debug=False):
    val0 = counts_to_pointwise_mutual_info (lexeme, class_term_ct,
                                                   class_non_term_ct, \
                                                   non_class_term_ct,
                                                   non_class_non_term_ct, \
                                                   debug=debug)
    try:
        return math.sqrt(val0)
    except ValueError:
        # neg log
        return - math.sqrt(-val0)
              

def counts_to_pointwise_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Per lexeme MI calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events.  The min of these 2
    counts is used in discounting (Pantel 2003 [Diss], Section  3.2, Eq. 3.7, p.50).
    """
    # print lexeme, class_term_ct,class_non_term_ct,non_class_term_ct,non_class_non_term_ct    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## discount computations
    discounting_meas = float(min(class_ct,term_ct))
    count_scaler = float(class_term_ct)/(class_term_ct + 1)
    discounting_scaler = discounting_meas/(discounting_meas + 1)
    discount = discounting_scaler * count_scaler

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)
    res = discount * pointwise_logprob
    return res

counts_to_hd_pointwise_mutual_info = counts_to_pointwise_mutual_info

def counts_to_normalized_pointwise_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Per lexeme MI calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events.  The min of these 2
    counts is used in discounting (Pantel 2003 [Diss], Section  3.2, Eq. 3.7, p.50).
    """
    # print lexeme, class_term_ct,class_non_term_ct,non_class_term_ct,non_class_non_term_ct    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## Total ct
    log_N = my_log(class_ct + non_class_ct)
    #### Probs
    # N.1
    class_logprob = my_log(class_ct) - log_N
    # N1.
    term_logprob = my_log(term_ct) - log_N
    # N11
    class_term_joint_logprob = my_log(class_term_ct) - log_N
    class_term_logprob_prod = class_logprob + term_logprob

    #pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)
    pointwise_logprob = (class_term_joint_logprob - class_term_logprob_prod)
    res = pointwise_logprob/(-class_term_joint_logprob)
    if res < 0:
        res = 0
    return res


def easy_counts_to_pointwise_mutual_info (class_term_ct, class_ct, term_ct, N,\
                                          debug=False):
    ## discount computations
    discounting_meas = float(min(class_ct,term_ct))
    count_scaler = float(class_term_ct)/(class_term_ct + 1)
    discounting_scaler = discounting_meas/(discounting_meas + 1)
    discount = discounting_scaler * count_scaler
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)
    res = discount * pointwise_logprob
    return res

def counts_to_shifted_pointwise_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Per lexeme MI calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events.  The min of these 2
    counts is used in discounting (Pantel 2003 [Diss], Section  3.2, Eq. 3.7, p.50).
    """
    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## discount computations
    discounting_meas = float(min(class_ct,term_ct))
    count_scaler = float(class_term_ct)/(class_term_ct + 1)
    discounting_scaler = discounting_meas/(discounting_meas + 1)
    discount = discounting_scaler * count_scaler

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    #pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)
    pointwise_logprob = my_log(1 + (class_term_joint_prob/class_term_prob_prod))
    res = discount * pointwise_logprob
    return res

def gr_counts_to_pointwise_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Per lexeme MI calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events.  The min of these 2
    counts is used in discounting (Pantel 2003 [Diss], Section  3.2, Eq. 3.7, p.50).
    """
    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## discount computations
    discounting_meas = float(min(class_ct,term_ct))
    count_scaler = float(class_term_ct)/(class_term_ct + 1)
    discounting_scaler = discounting_meas/(discounting_meas + 1)
    discount = discounting_scaler * count_scaler

    ## Total ct
    log_N = grefenstettes_log(class_ct + non_class_ct)
    #### Probs
    # N.1
    class_log_prob = grefenstettes_log(class_ct) - log_N
    # N1.
    term_log_prob = grefenstettes_log(term_ct) - log_N
    # N11
    class_term_log_joint_prob = grefenstettes_log(class_term_ct) - log_N 
    class_term_log_prob_prod = class_log_prob + term_log_prob

    pointwise_logprob = class_term_log_joint_prob - class_term_log_prob_prod
    res = discount * pointwise_logprob
    return res


def counts_to_undiscounted_pointwise_mutual_info (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Per lexeme MI calculation.  This version leaves out Pantel's discounting factor,
    which counteracts PMI's known tendency to overvalue rare events.  For experimenting
    to see how much discounting improves PMI.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events.
    """
    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## discount computations
    # discounting_meas = float(min(class_ct,term_ct))
    # count_scaler = float(class_term_ct)/(class_term_ct + 1)
    # discounting_scaler = discounting_meas/(discounting_meas + 1)
    # discount = discounting_scaler * count_scaler

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)
    # res = discount * pointwise_logprob
    return pointwise_logprob


def counts_to_log_pmi (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    divide pmi by log,  a form of normalization.
    """

    res = counts_to_pointwise_mutual_info(lexeme, class_term_ct, class_non_term_ct, \
                                          non_class_term_ct, non_class_non_term_ct, \
                                          debug=debug)
    return res/math.log(class_term_ct + non_class_term_ct + 1)

def counts_to_mutual_info_nonneg (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Leave out neg values of counts to pmi
    """

    res = counts_to_mutual_info(lexeme, class_term_ct, class_non_term_ct, \
                                          non_class_term_ct, non_class_non_term_ct, \
                                          debug=debug)
    if res > 0:
        return res
    else:
        return 0

def counts_to_pointwise_mutual_info_nonneg (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     debug=False):
    """
    Leave out neg values of counts to pmi
    """

    res = counts_to_pointwise_mutual_info(lexeme, class_term_ct, class_non_term_ct, \
                                          non_class_term_ct, non_class_non_term_ct, \
                                          debug=debug)
    if res > 0:
        return res
    else:
        return 0
    


def counts_to_lin_pointwise_mutual_info(lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, \
                                     lin_cond_deprel_prob,debug=False):
    #class_ct = class_term_ct + class_non_term_ct
    #non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct
    non_term_ct = class_non_term_ct + non_class_non_term_ct

    ## Total ct
    #N = class_ct + non_class_ct
    N = term_ct + non_term_ct
    #### Probs
    # N.1
    #class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (term_prob * lin_cond_deprel_prob)

    pointwise_logprob = my_log(class_term_joint_prob/class_term_prob_prod)

    return pointwise_logprob


def counts_to_t_test (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    For legacy reasons I will continue to call this t_test.  It is actually
    what is called z-score in Smadja 1993, Fontenelle et al.1994 and Scone
    and Jurafsky 2001. This (mis-)naming follows Curran but I remain
    responsible for my own mistakes.

    Per lexeme t_test calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events. 
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)
    independence_divergence = class_term_joint_prob - class_term_prob_prod
    
    res = independence_divergence/(class_term_prob_prod**0.5)

    return res

def counts_to_t_test_rescaled (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    For legacy reasons I will continue to call this t_test.  It is actually
    what is called z-score in Smadja 1993, Fontenelle et al.1994 and Scone
    and Jurafsky 2001. This (mis-)naming follows Curran but I remain
    responsible for my own mistakes.

    Per lexeme t_test calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events. 
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)
    independence_divergence = class_term_joint_prob - class_term_prob_prod
    
    res = independence_divergence/(class_term_prob_prod**0.5)

    return pow(N,0.5) * res

def counts_to_scp (lexeme, class_term_ct, class_non_term_ct, \
                   non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    symmetric conditional prob:  Ferreira & Pereira 1999.

    P_{xy}^2/(P_{x}*P_{y})
    This equals P(x|y)*P(y|x)
    """
    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct
    N = class_ct + non_class_ct

    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    return (class_term_joint_prob **2)/(class_prob * term_prob)

def counts_to_log_scp (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    symmetric conditional prob:  Ferreira & Pereira 1999.

    log(P_{xy}^2/(P_{x}P_{y}))
    This equals log(P(x|y)*P(y|x)_

    Note that what was actually implemented for the min sim functions
    was log(1 + P(x|y)P(y|x)).
    This worked cpbosiderably for the min than sim(-v1,-v2)
    """
    
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct
    N = class_ct + non_class_ct

    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)

    #return (2 * my_log(class_term_joint_prob)) - (my_log(class_prob) + my_log(term_prob))
    return my_log(1 + ((class_term_joint_prob**2)/(class_prob *term_prob)))

def counts_to_dice (lexeme, class_term_ct, class_non_term_ct, \
                    non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    dice as a weight function.
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    return (2 * float(class_term_ct))/(class_ct + term_ct)

def counts_to_log_dice (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    log of dice as a weight function.
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    return my_log(2 * class_term_ct) - my_log(class_ct + term_ct)

def counts_to_z_score (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    For legacy reasons I will continue to call this z_score.  It is actually
    t-test as used in Church and Hanks 1990 and presented in Manning & SChuetze.
    
    Per lexeme z_score calculation.
    
    Returning a number x s/t real z-score = (N ** .5) * x.
    Since we only ever do colmparisons of linear combos z scores this
    constant factor does not matter.
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)
    independence_divergence = class_term_joint_prob - class_term_prob_prod
    
    res = independence_divergence/((class_term_joint_prob)**0.5)

    return res

def counts_to_t_test_discounted (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Per lexeme t_test calculation.
    
    4 counts used.  All are assumed to be of disjoint sets, the counts
    of the C_T, C_non_T, non_C_T, and non_C_non_T events.  Since the 4 possible
    outcomes for binary random variables T and C are exhausted,
    the sum of all 4 counts is the total number of events, called N. The sum of
    C_T and non_C_T is the total number of T events, and the
    sum of C_T and C_non_T is the total number of C events. 
    """
    
    global N, class_ct, non_class_ct, term_ct
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct

    term_ct = class_term_ct + non_class_term_ct

    ## discount computations
    discounting_meas = float(min(class_ct,term_ct))
    count_scaler = float(class_term_ct)/(class_term_ct + 1)
    discounting_scaler = discounting_meas/(discounting_meas + 1)
    discount = discounting_scaler * count_scaler

    ## Total ct
    N = class_ct + non_class_ct
    #### Probs
    # N.1
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    # N11
    class_term_joint_prob = float(class_term_ct)/N
    class_term_prob_prod = (class_prob * term_prob)
    independence_divergence = class_term_joint_prob - class_term_prob_prod
    
    res = independence_divergence/(class_term_prob_prod**0.5)

    return discount * res

def counts_to_t_test_nonneg (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Per lexeme t_test calculation.

    This variant never returns a negative value.  Following Curran following Hindle. This
    is ACTUALLY what Curran calls t-test.
    """

    res = counts_to_t_test (lexeme, class_term_ct, class_non_term_ct, \
                            non_class_term_ct, non_class_non_term_ct, debug=debug)
    if res > 0:
        return res
    else:
        return 0

def counts_to_t_test_shifted (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Shift t-test vals to the right to grab some points of significant info.

    [-.0005,0] is the desired interval of neg vals to capture.  Map those into
    the region [0,.00005]. Shift all positive values accordingly.
    """

    res0 = counts_to_t_test (lexeme, class_term_ct, class_non_term_ct, \
                            non_class_term_ct, non_class_non_term_ct, debug=debug)
    shift = .0005
    scale = 1
    shift = shift * scale
    if res0 > 0:
        return res0 + shift
    elif res0 < 0:
        return (res0 * scale) + shift
    else:
        return 0

def counts_to_t_test_shifted_discounted (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Shift t-test vals to the right to grab some points of significant info.

    [-.0005,0] is the desired interval of neg vals to capture.  
    """

    res0 = counts_to_t_test_discounted (lexeme, class_term_ct, class_non_term_ct, \
                                        non_class_term_ct, non_class_non_term_ct, debug=debug)
    shift = .0005
    scale = 1
    shift = shift * scale
    if res0 > 0:
        return res0 + shift
    elif res0 < 0:
        return (res0 * scale) + shift
    else:
        return 0

def counts_to_t_test_scale_shifted (lexeme, class_term_ct, class_non_term_ct, \
                                     non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Shift t-test vals to the right to grab some points of significant info.

    [-.0005,0] is the desired interval of neg vals to capture.  Map those into
    the region [0,.00005]. Shift all positive values accordingly.
    """

    res0 = counts_to_t_test (lexeme, class_term_ct, class_non_term_ct, \
                            non_class_term_ct, non_class_non_term_ct, debug=debug)
    shift = .0005
    scale = .1
    scaled_shift = shift * scale
    if res0 > 0:
        return res0 + scaled_shift
    elif res0 < 0:
        return (res0 * scale) + scaled_shift
    else:
        return 0

def counts_to_pointwise_poisson_info (lexeme, class_term_ct, class_non_term_ct, \
                                      non_class_term_ct, non_class_non_term_ct, debug=False):

    """
    JMG:  Fixing this.  lambda was assumed to be:

    p(term) * p(class) * n
    where n is C{term_ct} (C{class_term_ct} + C{non_class_term_ct}).

    Shd be

    p(class) * n [were doing p(k | term) using term_ct as the toal numnber of tries]

    Also  we were comparing  prob(class) to prob(k instances of class),
    and that made no sense. 
    
    Inspired by Church and Hanks "Inverse Document Frequency (IDF): A Measure of
    Deviations from Poisson" => "A good keyword is far from Poisson".  They answer the
    question "Why does Poisson, a function of raw frequency,
    systemtically OVERESTIMATE the doc freq of good keywords?"

    ***** MAY NOT BE CORRECT! *********
    This is SIGNIFICANTLY better (run_ukb eval) with negative scores thrown away.
    Suggests that this this Poisson test is a pretty good significance test.  But
    still not competitive on the nearest neighbor task.  At all ranks underperforms
    ordinary PMI (with dot min).
    ***** MAY NOPT BE CORRECT! *********

    k is the number of occurrences of the dependendent with the head hence C{class_term_ct}.
    Assume k  >= 10 and we can use Stirling's formula to appximate k! [k * log k =~ log k!]

    lambda is the expected number of occurrences of class in n tries.
    Assuming  indepneendenc, that's p(class) * n
    n is C{term_ct} (C{class_term_ct} + C{non_class_term_ct}).

    Poisson_prob(k,lambda) = 1/k! * (lamdbda**k) * (e**(-lambda))
                           = pow(factorial(k),-1) * pow(lambda,k) * math.exp(-lambda)
                           = pow(factorial(class_term_ct),-1) * pow(lambda,class_term_ct) * math.exp(-lambda)
    where lambda = p(class) * C{term_ct}

    we do a pmi style comparison.  log (expected_p/actual p)

    We use the mean of the poisson dist poisson(lambda,lambda) [prob of lambda occurrences]

    lambda = expected # of occurrences of class = class_prob * term_ct

    We use Sterling's approx of ln n!: ln k! = k ln k.

    pois(lambda,lambda)/pois(k,lambda) = k!/lambda! * lambda^{lambda-k} * e^{k-lambda}

    The more improbable k occurrences, the more significant, so we put
    the k event in the denominator. If k is greater than lambda, this
    ratio is > 1, the log of it is positive.

    on the log scale:
           log k! - log lambda! + (lambda - k) log lambda + k- lambda

    Now consider a gross approx of k! using the first term of sterling
    we'll call gross sterling (not valid for small numbers)
    On the log scale, using gross sterling:

    k ln k - lambda ln lambda + (lambda - k) ln lambda + k -lambda
    k ln k - lambda ln lambda + lambda ln lambda - k ln lambda + k - lambda
    k ln k - k ln lambda + k - lambda
    k (ln k - ln lambda + 1) - lambda

    Complaint about quuasthoff wolff version.  Not properly scaled
    around zero. [This seems to matter]

    Tried this with gross sterling.  Tried it with true factyorial for low k
    and sterling with  one more term for low k.

    """
    ## The two row counts
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct
    ## The two col counts
    term_ct = class_term_ct + non_class_term_ct
    #non_term_ct = class_non_term_ct + non_class_non_term_ct
    N = class_ct + non_class_ct
    class_prob = float(class_ct)/N
    
    ## Old version
    ##log_class_term_joint_prob = math.log(class_term_ct) - math.log(term_ct)
    ## poisson_logprob = log_poisson (class_term_ct, class_ct,term_ct,N)

    ## res = log_class_term_joint_prob - poisson_logprob
    ## Total ct

    k = class_term_ct
    # N1.
    lambd = term_ct * class_prob
    lambda_delta = lambd - k
    log_lambda = math.log(lambd)
    # gross sterling
    #res = k * (math.log(k) - log_lambda +1) - lambd
    res = log_factorial_approx(k) - log_factorial_approx(lambd) + \
          lambda_delta * (math.log(lambd) - 1)
    if res < 0:
        res = 0
    return res


def counts_to_qw_poisson (lexeme, class_term_ct, class_non_term_ct, \
                           non_class_term_ct, non_class_non_term_ct, debug=False):
    """
    Approximation of laplace significance measure in Quasthoff & Wolff (2002)[Q&W]

    (lambd - (k * log_lambda) + math.lgamma(k))/math.log(term_ct)

    
    The particular approximation of ln k! used here is lgamma.
    A version using stirl;ing's approximation comes from Bordag (2008).
    This performs best on Bordag's evaluation of co-occurrence significance 
    measures

    Note:  All of Bordag's evaluation results favor measures that favor large counts.
    Q& W ays "This probability approx good when (k + 1)/lambda > 10."  Since lambda is
    the expected count  given prob of w1, w2, this means the approximation is good  for counts WAY higher
    than the xpected count.  Since k has a minimum of 1, we're getting a good approx whenever
    lambda < .05 (typically the case).  Otherwise we're using the first term of an infinite series
    ALL of whose terms are positive, so we're underestimating the true prob, and therefore OVER
    estimating the significance.  A reasanble refinement is to use
    more terms of the series when lambda >= .05.

    Note: THis was originally misimplemented.  Note that k is NOT a cooccurreence count it's
    count of the number of times class was a dependent of term (not bidirectional),
    so lambda in FACT is:
      class_prob * term_ct (we had term_ct tries to produce an outcome with prob class_prob)
    We had lambda  the way Q&W did as
      class_prob * term_prob * class_term_ct (we had term_ct tries to produce an outcome with prob class_prob)
    This is right when k is a true co-occurrence ct but wrong when it is
    a dependency ct.

    Note this is an approximation that MIGHT be improved by removing
    the dependency on stirling's approximation.

    Note this CANNOT be negative (neg log of prob divided by an x > 1).

    An alternative version of the same idea is is just to use the -log binomial(x>=k;lambda)/log n.
    Why is poisson better than binomial here?  Alternatively, multinomial?

    Another alternative.  Foloowing Q&W we divide by log n.  This divides very long
    vectores by a big num, disadvantaging them, apparently.  But if pois(k,lambd)
    generally gets bigger for longer n (more tries),  this makes -log pois(k,lambd) get
    smaller (less information/surprise).  Prediction long attractors should be rare.
    Danger:  The cosine pitfall.  THe spector of short attractors.
    """
    k = class_term_ct
    # A form of smoothing.  Guarantees no zero div error ewhen we divide by log(term_ct)
    # below because term_ct always greater than 1 (sinc emin val pf term_class_ct = 1).
    # Uniformly increases the value of SOME non class feature by 1, woithout choosing which one.
    non_class_term_ct += 1
    ## The two row counts
    class_ct = class_term_ct + class_non_term_ct
    non_class_ct = non_class_term_ct + non_class_non_term_ct
    ## The two col counts
    term_ct = class_term_ct + non_class_term_ct
    non_term_ct = class_non_term_ct + non_class_non_term_ct

    ## Total ct
    N = class_ct + non_class_ct
    val_list = [N, class_ct, non_class_ct, term_ct, non_term_ct, class_term_ct, \
                class_non_term_ct,non_class_term_ct, non_class_non_term_ct]
    ## lambda = n p(class) * p(term)
    class_prob = float(class_ct)/N
    # N1.
    term_prob = float(term_ct)/N
    #######  lambda. Expected freq of class
    lambd = term_ct * class_prob
    log_lambda = math.log(lambd)
    ## The next (erroneous) version pretty much comes out equiv to PMI, reduced by 1,
    ## to filter low values with neg filtering.  As one might expect, does
    ## very nearly as well as PMI (but not quite!).
    ## ln k - ln lambda - 1
    #res = math.log(class_term_ct) - log_lambda - 1
    try:
        log_k_fact = math.lgamma(k)
    except Exception, e:
        print 'lgamma error', k
        raise e
    try:
        res = (lambd - (k * log_lambda) + log_k_fact)/math.log(term_ct)
    except ZeroDivisionError, e:
        print 'Zero Div', lambd, k, log_lambda, term_ct, class_term_ct
        raise e
    return res
    
def log_poisson_s (class_term_ct, class_ct,term_ct,N):
    """
    Via Stirling's approx: log k! = (k * math.log(k)) - k + (0.5 * math.log(2 * math.pi * k))
    """
    k = class_term_ct
    ##  Try one: Expected number of class tokens in term_ct terms = p(class) * term_ct
    ##  works better than try two, less correct?
    lam = float(class_ct * term_ct)/N
    ##  Try two: Prob of class token = p(class)
    ##lam = float(class_ct)/N
    # log (1/k! * lambda^k * exp(-lambda)
    return -log_stirling(k) + (k * math.log(lam)) - lam
    #return -math.log(math.factorial(k)) + (k * math.log(lam)) - lam
    
def log_poisson (class_term_ct, class_ct,term_ct,N):
    """
    Via Stirling's approx: log k! = (k * math.log(k)) - k + (0.5 * math.log(2 * math.pi * k))
    """
    k = class_term_ct
    ##  Try one: Expected number of class tokens in term_ct terms = p(class) * term_ct
    ##  works better than try two, less correct?
    lam = float(class_ct * term_ct)/N
    ##  Try two: Prob of class token = p(class)
    ##lam = float(class_ct)/N
    # log (1/k! * lambda^k * exp(-lambda)
    if k > 5:
        return -log_stirling(k) + (k * math.log(lam)) - lam
    else:
        return -math.log(math.factorial(k)) + (k * math.log(lam)) - lam

def log_factorial_approx (k,thresh=10):
    """
    Using the lgamma fn inSTEAD of stirlings approx for log k!

    Shd be more accurate, especially for low integral values of k,
    where it is exact.
    """
    try:
        return math.lgamma(k+1)
    except OverflowError,e:
        print 'log factorial error: %.7f' % (k,)
        raise e

def log_stirling (k):
    return (k * math.log(k)) - k + (0.5 * math.log(2 * math.pi * k))

def my_log(num):
    """
    A workable strategy is to add return 0 for 0.
    """
    try:
        if abs(num) < 1.0000000e-108:
            return 0
        else:
            return math.log(num,2)
    except ValueError,e:
        print 'Cant take log of %.6f' % (num,)
        raise e


def grefenstettes_log(num):
    """
    A more principled strategy is to add 1 to all counts and take the log.

    Shd this be applied only to counts o can it be applied to probs?
    """
    return math.log(num + 1,2)


def sqrt_freq  (this_dict, sum_cts):
    """
    Weight/normalizing function used for Hellinger distance.  

    Dot prod can then be used as similarity function.
    """
    norm = pow(sum_cts, 0.5)
    for deprel in this_dict:
        val = this_dict[deprel]
        this_dict[deprel] = (pow(val,0.5)/norm)

def conditional_probs (this_dict, sum_cts):
    for deprel in this_dict:
        val = this_dict[deprel]
        this_dict[deprel] = (float(val)/sum_cts)

####################################################################################
####################################################################################
##
##     Similarity functions (v1, v2)  ==>  RealNumber
##
####################################################################################
####################################################################################


def IRad (vec1, vec2):
    """
    Implementation of the information radius measure
    evaluated in Dagan et al. 1997b.  See also Kai and
    Rijsbergen (2008[Sibson 1967])

    Lee (1997) manipulates this algebraically so that
    only iteration over shared attributes is necessary.
    Look this up.  This implementation obviously
    not taking advantage.
    """

    Dp_pqhat = 0
    for deprel in vec1:
        Dp_pqhat += IRad_helper(deprel,vec1,vec2)
    for deprel in vec2:
        Dp_pqhat += IRad_helper(deprel,vec2,vec1)
    return Dp_pqhat

def IRad_helper(deprel,vec1,vec2):
    p = vec1[deprel]
    q = vec2.get(deprel,None)
    # 
    if q:
        return p * (math.log(p/(p+q),2) + 1.0) # p * log (2p/(p+q)) log base 2
    else:
        return p

def AsymIRad (vec1, vec2):
    """
    Implementation of HALF the information radius measure
    evaluated in Dagan et al. 1997b.  See also Cai and
    Van Rijsbergen (2007[Sibson 1967])

    This version is no longer symmetric, and is therefore much more efficient.
    But when vec1 comes from a freq word and vec2 from a rare word,
    will it pay too much attention to vec2's weightings?

    To be played with before trying an experiment.
    Note: Compare this with Lee's alpha-skew, also an asymmetric
    version of IRad.
    """
    Dp_pqhat = 0
    for deprel in vec2:
        Dp_pqhat += IRad_helper(deprel,vec2,vec1)
    return Dp_pqhat

def lin (vec1, vec2, alpha=.5):
    """
    Just the formula.  Ignores the possibility of negative feat values.

    Prblem with this: When val1 and val2 are positive & negative, repsectively,
    they are treated as a commonality.  When both are negative they
    subtract an arbitrary amount from the commonality and the total,
    possibly raising, possiblt lowering the result.  It tends to hurt
    high scores, which it certainly shdnt.
    7/28 Fixed major logical bug!
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        val1 = vec1.get(deprel,0.0)
        if val1:
            common += val2 + val1
    return common/((alpha* vec2_vals) + (1-alpha)*sum(vec1.itervalues()))

def dice_dot_prod (vec1, vec2):
    """
    This is what the data mining community calls dice coefficient,
    reduces to dice set with binary vector.

    Alternative derivatiion: If we look at Dice as computing
    the information shared by v1i and v2i using an operation Op
    and Op is addition (v1i + v2i),  we get lin;  if Op is min,
    we get dice_dagger,   if it is product (v1i * v2i), we get this.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += (val2 ** 2)
        val1 = vec1.get(deprel,0.0)
        if val1:
            common += val2 * val1
    return (2 * common)/(vec2_vals + sum(x**2 for x in vec1.itervalues()))

def dice_dot_prod_alpha (vec1, vec2, alpha=.3):
    """
    We start with 1/(e^2) as the default alpha, then try .3.

    Parameterized version of ddp in which 0 =< alpha =< 2.0;
    alpha determines the weighting of recall of w1-feats by w2
    versus precision of w1-feats by w2 (recall= sigma(w1,w2)/sigma(w_1,w_1),
    precision=sigma(w1,w2)/sigma(w_2,w_2)).  alpha = 1.0 is even
    weighting (equivalent to default ddp), alpha = 1.5 weights recall
    over precision.  Hypoithesized to work better with lower rank w1's.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += (val2 ** 2)
        val1 = vec1.get(deprel,0.0)
        if val1:
            common += val2 * val1
    return (2 * common)/((alpha * vec2_vals) + ((2-alpha) *sum(x**2 for x in vec1.itervalues())))

def jaccard_dot_prod (vec1, vec2):
    """
    VERY like dice dot prod.  This is what data mining community
    uses for jacc and jacc set ( vectors binary => |intersection|/|union|)
    and dice dot prod is what they use for dice coefficieint.

    Seems to find all the same pairs as dice dot prod.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        val1 = vec1.get(deprel,0.0)
        vec2_vals += (val2 ** 2)
        if val1:
            common += val2 * val1
    return common/(vec2_vals - common + sum(x**2 for x in vec1.itervalues()))

def dice_dot_min (vec1, vec2):
    """
    Like dice_dot_product, except we use the max shared info in
    the normalization factor. So no factor 2 in  the numerator
    """
    common = 0.0
    vec2_vals = 0.0
    vec1_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        val2_sq = (val2 ** 2)
        vec2_vals += val2_sq
        val1 = vec1.get(deprel,0.0)
        if val1:
            common += val2 * val1
            if val1 <= val2:
                vec1_vals -= (val1**2)
            else:
                vec2_vals -= val2_sq                
    return common/(vec1_vals + vec2_vals + sum(x**2 for x in vec1.itervalues()))

def dice_dot_prod_sq_rt (vec1, vec2):
    """
    If we look at Dice as using operation on v1i v2i and we do v1i + v2i
    we get lin, if we do min we get dice, if we do * we get this or dice_dot_prod. If we do geometric mean we get this.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        val1 = vec1.get(deprel,0.0)
        if val1:
            prod1 = val2 * val1
            if prod1 > 0:
                        common += pow(prod1,0.5)
    return (2 * common)/(vec2_vals + sum(x for x in vec1.itervalues()))

def dice_dot_prod_sq_rt_all_neg (vec1, vec2):
    """
    If we look at Dice as using operation on v1i v2i and we do v1i + v2i
    we get lin, if we do min we get dice, if we do * we get this or dice_dot_prod. If we do geometric mean we get this.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        val1 = vec1.get(deprel,0.0)
        if val1:
            prod1 = val2 * val1
            if prod1 > 0:
                        common += pow(prod1,0.5)
    return - (2 * common)/(vec2_vals + sum(x for x in vec1.itervalues()))

def dot_prod_sq_rt (vec1, vec2):
    """
    dot_x fn in which the SIS is geometric mean.  THis is a convenience.
    This shd be implemented with the square roots taken in the weighting scheme
    and dot_prod used as the sim function.
    """
    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        val1 = vec1.get(deprel,0.0)
        if val1:
            prod1 = val2 * val1
            if prod1 > 0:
                common += pow(prod1,0.5)
    return common

def dice_dot_min_sq_rt (vec1, vec2):
    """
    Like dice_dot_sq_rt, except we use the max shared info in
    he normalization factor. So no factor 2 in  the numerator
    """
    common = 0.0
    vec2_vals = 0.0
    vec1_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        val1 = vec1.get(deprel,0.0)
        vec2_vals += val2
        if val1:
            prod1 = val2 * val1
            if prod1 > 0:
                        common += pow(prod1,0.5)
            if val2 <= val1:
                # Max only in normalization. Min canceled out
                vec2_vals -= val2
            else:
                # Max only in normalization. Min canceled out
                vec1_vals -= val1
    return common/(vec1_vals + vec2_vals + sum(x for x in vec1.itervalues()))

def lin_no_neg (vec1, vec2):

    """
    A version of lin that works!

    Following Hindle: If both feature vals are positive
    or both are negative, add them to the commonality.
    Else ignore.  Huge difference in performance when
    looking at nearest neighbors, because very common words are
    otherwise hard to beat.
    """

    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        val1 = vec1.get(deprel,0.0)
        if val1 > 0 and val2 > 0:
            common += val2 + val1
        elif val1 < 0 and val2 < 0:
            common += val2 + val1
            # Problem with this: Must do the same in the denominator
            # Intuition: This case counts POSITIVELY toward what is common, which suggests
            # common += -val2 + -val1
    return common/(vec2_vals + sum(vec1.itervalues()))

def lin_no_neg_principled (vec1, vec2):

    """
    A "more principled" less efficient  version of
    lin that does not work.

    Following Hindle: If both feature vals are positive
    or both are negative, add them to the commonality.
    In this case we add positive values to the commonality
    when both are negative, and positive values to the totality
    when both are negative.
    
    Else ignore.  Huge difference in performance when
    looking at nearest neighbors, because very common words are
    otherwise hard to beat.

    This performns terribly. Very common words have a huge advantage.
    And normalization does not solve this!
    """

    common = 0.0
    other_total_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        val1 = vec1.get(deprel,0.0)
        if val1 > 0 and val2 > 0:
            other_total_vals += val2
            common += val2 + val1
        elif val1 < 0 and val2 < 0:
            mag = abs(val1 + val2)
            ## Val1 vals needed in other_total_vals because otherwise no neg val1s are summed into denominator
            other_total_vals += mag
            common += mag
    return common/(other_total_vals + sum((v for v in vec1.itervalues() if v > 0)))



def lin_no_neg2 (vec1, vec2):

    """
    The idea: Ignore ALL negative feature vals in common.  But note
    they are still making a penalty contribution in the denominator,
    raising the relative importance of shared positive feats.

    This means it is possible that lin_no_neg2(v1,v1) > 1.0
    """

    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        if val2 > 0:
            val1 = vec1.get(deprel,0.0)
            if val1 > 0:
                common += val2 + val1
    return common/(vec2_vals + sum(vec1.itervalues()))

def lin_no_neg3 (vec1, vec2):

    """
    Variant of lin_no_neg2 with penalties.
    """

    common = 0.0
    vec2_vals = 0.0
    il = vec2.iteritems()
    for (deprel,val2) in il:
        vec2_vals += val2
        val1 = vec1.get(deprel,0.0)
        if val1 > 0 and val2 > 0:
            common += val2 + val1
        # Penalize!
        elif val1 > 0 and val2 < 0:
            common += val2
        # Penalize!
        elif val1 < 0 and val2 > 0:
            common += val1
    return common/(vec2_vals + sum(vec1.itervalues()))

def lin_sim_score (lexeme1, lexeme2, n=500):
    wv1 = word_deprel_ct_dict[lexeme1].items()
    wv2 = word_deprel_ct_dict[lexeme2].items()
    wv1.sort(key=lambda x: x[1],reverse=True)
    wv2.sort(key=lambda x: x[1],reverse=True)
    nwv1 = {}
    nwv2 = {}
    for (k,v) in wv1[:n]:
        nwv1[k] = v
    for (k,v) in wv2[:n]:
        nwv2[k] = v
    return lin(nwv1, nwv2)

def run_lin_sim_neighborhood_experiment (nn_dict,center = ('wanted','VBD'), n=500):
    """
    Using C{lin}, results are basically awful with full vectors. As suggested in Zhitomirsky-Geffet
    and Dagan 2009. "is" is the nn of "wants" and a number of other verbs.
    Pruning helps somewhat [get rid of all but the n top-valued feats], so this is what the
    parameter n of this experiement function does.  But results still fall far short
    wat can be gotten through cos, pmi, using Pantel's formulation.

    Answer: use a nonneg version of lin.
    """
    results = {}
    for ol in nn_dict:
        results[ol] = lin_sim_score(center,ol,n=n)
    res_list = results.items()
    res_list.sort(key=lambda x: x[1], reverse=True)
    for (k,v) in res_list:
        print k, v
#
# def IRad (vec1, vec2):
#     """
#     Implementation of the information radius measure
#     of Dagan et al. 1997b
#     """
#     Dp_pqhat = 0
#     for deprel in vec1:
#         Dp_pqhat += IRad_helper(deprel,vec1,vec2)
#     for deprel in vec2:
#         Dp_pqhat += IRad_helper(deprel,vec2,vec1)
#     return Dp_pqhat
# 
# def IRad_helper(deprel,vec1,vec2):
#     p = vec1[deprel]
#     logp = math.log(p)
#     q = vec2.get(deprel,0)
#     pqhat = math.log(p+q) - 1 # log((p+q)/2)
#     return p * (logp - pqhat)
#     
# 
            
def dot_prod (vec1, vec2):
    """
    Silly sparse vector implementation of dot product.

    vec1 and vec2 are actually dictionaries with integer keys
    and float vals, representing sparse vectors.  Absence of
    key k implies vec[k] = 0.

    A;t From itertools (for list vectors of eq len):
    return sum(imap(operator.mul, vec1, vec2))
    """
    
    res = 0.0
    for (k,v) in vec2.iteritems():
        res += (v * vec1.get(k,0))
    return res

def dot_sum (vec1, vec2):
    """
    Silly sparse vector implementation of dot sum

    vec1 and vec2 are actually dictionaries with integer keys
    and float vals, representing sparse vectors.  Absence of
    key k implies vec[k] = 0.
    """
    res = 0.0
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res += (v1 + v2)
    return res

def dot_sum_all_neg (vec1, vec2):
    """
    Silly sparse vector implementation of dot sum

    vec1 and vec2 are actually dictionaries with integer keys
    and float vals, representing sparse vectors.  Absence of
    key k implies vec[k] = 0.
    """
    res = 0.0
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res -= (v1 + v2)
    return res

def dot_min (vec1, vec2):
    """
    Silly sparse vector implementation of dot min

    .688 0-500
    """
    res = 0.0
    #res = neg_infinity
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res += min(v1, v2)
    return res

def dot_min_union (vec1, vec2):
    """
    normalize dot min by dividing by the caridality of the
    union of the two feature sets.
    """
    res = 0.0
    vec2_ctr = 0
    vec1_ctr = len(vec1)
    for (k,v2) in vec2.iteritems():
        vec2_ctr += 1
        v1 = vec1.get(k,0)
        if v1:
            vec1_ctr -= 1
            res += min(v1, v2)
    return res/(vec1_ctr + vec2_ctr)


def dot_min_all_neg (vec1, vec2):
    """
    for log prob wt schemes: always negative valued
    """
    res = 0.0
    #res = neg_infinity
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res -= max(v1, v2)
    return res


def exp_dot_min_all_neg (vec1, vec2):
    """
    for log prob wt schemes: always negative valued

    This one makes less sense, works less well!
    """
    res = 0.0
    #res = neg_infinity
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res -= min(v1, v2)
    return res


def dot_min_naive (vec1, vec2):
    """
    Silly sparse vector implementation of dot min.

    This variant will always subtract given a min value.
    Prediction: Slight help for high rank hurts for low.
    Validated:   .689 for high rank set (0-500) vs .688 for dot_min
    and .488 for low vs. 510 for dot min when iteraing through vec2
    vals.  ... when iterating throughvec1 vals.
    """
    res = 0.0
    for (k,v1) in vec1.iteritems():
        res += min(vec2.get(k,0), v1)
    return res

def dot_min_fussy (vec1, vec2):
    """
    Silly sparse vector implementation of dot min.

    The cases of v1 or v2 < 0 are commented below. If we
    add 0 in all three cases, we have a nonneg value system.
    If we add mval (penalize) when v1 or v2 < 0 and add 0
    when both are, we have a compromise between an all neg system
    and a Hindle system, which adds 0 when v1 or v2 < 0
    and adds mval when both are.
                v1 or v2 < 0      Both < 0
                ============      ========
    .688 .509    minval             minval   (a     
    .677 .502    minval             -minval  (b     # cf. a penalty -> reward hurts.
    .683 .509    minval             0        (c
    .672 .508    0                  -maxval  (d     # cf. a,b Reward reduc hurts high rank more!
                                                    # Makes no sense.
    .673 .502    minval              -maxval (e
    .673 .511    0                   0       (f     # prediction <.683,.509
    .686 .507    0                   minval  (g 
    ## penalizing for ANY neg val works acording to the numbers above.
    ## Whic is why the first model is the best
                 0                   maxval
                 max(-maxval,minval) maxval
    .671  .497   avg(v1,v2)          maxval
    .684  .516  -min(|v1|,|v2|)     maxval
    A truly fussy alternative: penalty for OR case is -min(abs(v1),abs(v2))
    """
    res = 0.0
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            if v1 > 0:
                # If v2 < 0 min(v1, v2) is a penalty, shd be 0?
                if v2 < 0:
                    # min(v1, v2)
                    res += min(v1,v2) #-min(abs(v1),abs(v2)) # (v1 + v2)/2.0
                else:
                    res += min(v1, v2)
            elif v2 < 0:
                # v1 < 0, v2 < 0. A true similarity
                # res -= mval
                # Alternative:  +0 here. +|mval| may overestimate significance.
                res += min(v1,v2) #max(v1,v2)
                #res -= max(v1,v2)
                #res += min(v1,v2)
            else:
                # v1 < 0, v2 > 0
                # always a penalty, cd be 0
                #res += min(v1, v2)
                res += min(v1,v2) # -min(abs(v1),abs(v2)) # (v1 + v2)/2.0
    return res

def dot_avg(vec1, vec2):
    """
    Silly sparse vector implementation of dot x that takes
    avg of two values and sums.  Monotonic with dot_sum and
    really of no use as a sim fn.
    """
    res = 0.0
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            res += (v1 + v2)/2.0
    return res

from math import sqrt

def dot_min_sqrt (vec1, vec2):
    """
    Silly sparse vector implementation of dot min sqrt
    """
    res = 0.0
    for (k,v2) in vec2.iteritems():
        v1 = vec1.get(k,0)
        if v1:
            min_v_sqrt = sqrt(abs(min(v1,v2)))
            if v1 < 0:
                #if v2 > 0:
                    res -= min_v_sqrt
            else:
                res += min_v_sqrt
    return res


def cos (vec1, vec2):
    """
    Just a convenience.  For efficiency cosine
    is generally done by doing Euclidean normalization in
    the weighting scheme and then using dot prod.
    """
    v1_len_sq = sum(x**2 for x in vec1.itervalues())
    v2_len_sq = sum(y**2 for y in vec2.itervalues())
    res = 0.0
    for (k,v) in vec2.iteritems():
        res += (v * vec1.get(k,0))
    return res/pow(v1_len_sq * v2_len_sq, 0.5)

def dot_prod_no_neg (vec1, vec2):
    """
    Principled treatment of
    neg,pos pairs. Note to treat neg values \& pos product values alike is
    inconsistent with a sim score in the interval [0,1], which is arguably 
    what you want.  This version sums up sims, ignores diffs.  But what this 
    means is that one big sim overrules a lot of diffs.  Much looser def.

    The analogous idea may work better with jaccard?
    """
    res = 0
    for (k,v) in vec2.iteritems():
        val = (v * vec1.get(k,0))
        if val > 0:
            res += val
    return res


def dot_prod_eff (vec1, vec2):
    """
    This looks to be about 50% than the old fashioned way.
    A constant speedup.  Nope.  Not always! Retire.

    Note both treat neg values \& pos values alike.  This is inconsistent
    with a sim score in the interval [0,1], which is arguably what you want.
    """
    return sum((v * vec1.get(k,0) for (k,v) in vec2.iteritems()))

def old_jaccard (vec1, vec2):
    """
    Didnt remember: assoc weights may be negative.
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            if v2 <= v1:
                shared += v2
                max_share += v1
            else:
                shared += v1
                max_share += v2
        else:
            max_share += v2
    for (k1,v1) in vec1.iteritems():
        v2 = vec2.get(k1,0)
        if v2:
            # This pair already taken care of
            pass
        #elif v1 > 0:
        else:
            max_share += v1
    return shared/max_share


def jaccard (vec1, vec2):
    """
    Assumes assoc weights may be negative.So: Not to be used with
    counts_to_t_test_nonneg.

    Slows this down another little lick.

    To be used with pmi, t_test, not pmi_nonneg, t_test_nonneg.
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            if v2 <= v1:
                shared += v2
                max_share += v1
            else:
                shared += v1
                max_share += v2
        elif v2 > 0:
            max_share += v2
    for (k1,v1) in vec1.iteritems():
        v2 = vec2.get(k1,0)
        if v2:
            # This pair already taken care of
            pass
        elif v1 > 0:
            max_share += v1
    return shared/max_share

def simple_jaccard (vec1, vec2):
    """
    Binary set intersection jaccard.  |S1 ^ S2|/|S1 v S2|. The original
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            shared += v2
            max_share += v1
        else:
            max_share += 1
    for (k1,v1) in vec1.iteritems():
        v2 = vec2.get(k1,0)
        if v2:
            # This pair already taken care of
            pass
        else:
            max_share += v1
    return shared/max_share


def dice_dagger_fussy (vec1, vec2):
    """
    Only positive vals count as shared.

    But neg vals go into max share, reducing its mass,
    enhancing the importance of positive vals.  
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            if v2 <= v1:
                if v2 > 0:
                    shared += v2
            else:
                if v1 > 0:
                    shared += v1
        max_share += v2
    max_share += sum([v for v in vec1.itervalues()])
    return (2 * shared)/max_share

def dice_dagger (vec1, vec2):
    """
    The dagger portion of the name comes from Curran.  Nearly identical
    to jaccard and produces all the same pairings of nearest neighbors.

    Slightly faster to compute. (slightly less 3 times faster)
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            if v2 <= v1:
                shared += v2
            else:
                shared += v1
        if v2 > 0:
            max_share += v2
    max_share += sum([v for v in vec1.itervalues() if v > 0])
    return (2 * shared)/max_share

def dice_dagger_no_neg (vec1, vec2):
    """
    This version takes advantage of an always-positive weight function
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        max_share += v2
        if v1:
            if v2 <= v1:
                shared += v2
            else:
                shared += v1
    max_share += sum([v for v in vec1.itervalues()])
    return (2 * shared)/max_share

def dice_dagger_all_neg (vec1, vec2):
    """
    When wts from 0 to 1 are mapped to a log scale, max_share and
    shared will both be negative and we want the shared ct to come
    from the val of  lesser absolute value, i.e., more value.
    
    So this version takes advantage of an always-negative weight function
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        max_share += v2
        if v1:
            if v2 <= v1:
                shared += v1
            else:
                shared += v2
    max_share += sum([v for v in vec1.itervalues()])
    return (2 * shared)/max_share

def dice_dagger_naive (vec1, vec2):
    """
    this is the same as dice_dagger_no_neg.  It doesnt know
    there CAN even be negative values.  It is is here because
    my automatic naming scheme needs this to have a different name.

    Intended to be used with a with function of unrestricted range,
    to show how bad the results are.
    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        max_share += v2
        if v1:
            if v2 <= v1:
                shared += v2
            else:
                shared += v1
    max_share += sum([v for v in vec1.itervalues()])
    return (2 * shared)/max_share

def jaccard_no_neg (vec1, vec2):
    """
    This version assumes an association weight can only be pos (e.g. mut info no neg,
    t-test_no_neg) INSTEAD OF mapping all neg vals to 0. I.E., Can be used with
    counts_to_t_test_nonneg.

    """
    shared = 0.0
    max_share = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            if v2 <= v1:
                shared += v2
                max_share += v1
            else:
                shared += v1
                max_share += v2
        else:
            max_share += v2
    for (k1,v1) in vec1.iteritems():
        v2 = vec2.get(k1,0)
        if v2:
            # This pair already taken care of
            pass
        else:
            max_share += v1
    return shared/max_share

def alt_jaccard (vec1, vec2):
    """
    This underperforms jaccard significantly accdg to Curran (2003)
    """
    res = 0
    norm = 0
    for (k2,v2) in vec2.iteritems():
        v1 = vec1.get(k2,0)
        if v1:
            res += v2 * v1
            norm += v2 + v1
        else:
            norm += v2
    for (k1,v1) in vec1.iteritems():
        v2 = vec2.get(k1,0)
        if v2:
            # This pair already taken care of
            pass
        else:
            norm += v1
    return res/norm

## IRad returns a "distance" score, not a similarity score
## 0 means identical, bigger means more different
#distance_measures = [IRad,(dot_prod,counts_to_t_test)]
distance_measures = [IRad]
#neg_domain_pairs = [(dot_min,counts_to_log_scp)]
neg_domain_pairs = [ ] 

def sim_func_attributes(sim_func,wt_func):
    if (sim_func,wt_func) in neg_domain_pairs:
        best_val = neg_infinity
        comp_rel = lambda x,y: x > y
    elif (sim_func in distance_measures) or ((sim_func,wt_func) in distance_measures):
        best_val = sys.maxint
        comp_rel = lambda x,y: x < y
    else:
        best_val = 0
        comp_rel = lambda x,y: x > y
    return (best_val, comp_rel)

# def sim_max_min(new_val,best_val,sim_func):
#    if sim_func == IRad:
#        return (new_val < best_val)
#    else:
#        return (new_val > best_val)

## Possible values for C{normalize} flag
norm_values = ['l1','l2','','hell']

def test_norm_constraint(mi_func,normalize):
    """
    mi_func=sqrt_freq is for  Hellinger distance.
    mi_func = conditional_prob is for Information Radius.
    """
    assert normalize.lower() in norm_values, 'Normalize must be one of %s!' % (norm_values,)
    if normalize == 'l1':
        assert (mi_func == conditional_probs) or (mi_func == sqrt_freq), \
           'Mi Func %s not compatible with L1 norm! \nUse option mi_func=conditional_probs or mi_func=sqrt_freq.' % (mi_func.__name__,)
    elif normalize == 'hell':
        assert mi_func == counts_to_pointwise_mutual_info_nonneg


if __name__ == '__main__':
    import os.path
    from dependency_DB.dependency_db import read_dependency_file
    from dependency_DB.dep_rel_info import get_deprel_info, print_deprel_profile, get_relatum_and_function,decode_deprel,get_relatum
    
    home = os.environ['HOME']
    dependency_db_dir = os.path.join(home,'Desktop','SWIDSAS','dep_data')

    ## Assumed to exist
    dependency_keys_file = os.path.join(dependency_db_dir,'dependency_DB.keys')
    dependency_train_file = os.path.join(dependency_db_dir,'dependency_DB.train')
    # Using pruned model, thresshold = 6
    
    print 'Reading in raw counts %s' % (dependency_train_file,)
    (word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
                read_dependency_file(dependency_train_file, count_file=True, \
                                                suffix = '', ext = '')


    get_deprel_info (dependency_keys_file)
    
    def compute_poisson_probs_for_lexeme(lexeme,word_deprel_ct_dict, word_ct_dict,\
                                             deprel_ct_dict, total_deprels):
            this_dict = word_deprel_ct_dict[lexeme]
            term_ct = word_ct_dict[lexeme]
            #fn_cts = get_function_cts (lexeme,word_deprel_ct_dict)
            ctr = 0
            sig_ctr = 0
            sorted_dep_l = sorted(this_dict.keys())
            for deprel in sorted_dep_l:
                ctr += 1
                class_term_ct = this_dict[deprel]
                class_ct = deprel_ct_dict[deprel]
                # fussier version
                # log_class_term_joint_prob = math.log(class_term_ct) - math.log(total_deprels)
                # looser version
                log_class_term_joint_prob = math.log(class_term_ct) - math.log(term_ct)
                poisson_logprob = log_poisson (class_term_ct, class_ct,term_ct,total_deprels)
                res = log_class_term_joint_prob - poisson_logprob
                if res < 0:
                    res = 0
                if res:
                    print 'k: %d' % (class_term_ct,)
                    sig_ctr += 1
                    decoded_dr = get_relatum(deprel,False)
                    print '%35s  %.8f %.8f' % (decoded_dr, poisson_logprob,log_class_term_joint_prob)
            print
            print '%d out of %d items were significant.' % (sig_ctr,ctr)
   
    def alt_compute_poisson_probs_for_lexeme(lexeme,word_deprel_ct_dict, word_ct_dict,\
                                            deprel_ct_dict, total_deprels):
        this_dict = word_deprel_ct_dict[lexeme]
        term_ct = word_ct_dict[lexeme]
        fn_cts = get_function_cts (lexeme,word_deprel_ct_dict)
        for deprel in this_dict.keys():
            class_term_ct = this_dict[deprel]
            class_ct = deprel_ct_dict[deprel]
            (d_fn,d_lex,d_pos) = decode_deprel(deprel)
            ##log_class_term_joint_prob1 = math.log(class_term_ct) - math.log(total_deprels)
            #log_class_term_joint_prob = math.log(class_term_ct) - math.log(term_ct)
            #pp = poisson (class_term_ct, class_ct,term_ct,total_deprels)
            dd_ct = fn_cts[d_fn]
            log_class_term_joint_prob = math.log(class_term_ct) - math.log(term_ct)
            pp = log_poisson (class_term_ct, class_ct,dd_ct,total_deprels)
            decoded_dr = get_relatum_and_function(deprel,False)
            if pp < log_class_term_joint_prob:
                print '%7s  %.8f %.8fx' % (decoded_dr, pp,log_class_term_joint_prob)


    def get_function_cts (lexeme, word_deprel_ct_dict):
        fn_cts = Counter()
        for (dc,ct) in word_deprel_ct_dict[lexeme].iteritems():
            (d_fn,d_lex,d_pos) = decode_deprel(dc)
            fn_cts[d_fn] += ct
        return fn_cts
        
    lexeme = ('church','NN')
    compute_poisson_probs_for_lexeme(lexeme,word_deprel_ct_dict, word_ct_dict,
                                      deprel_ct_dict, total_deprels)


